# tag: ewok authorisation module

package Ewok::Authz;

require Exporter;
use Data::Dumper;
use Carp;
use Ewok::Config;
use Ewok::Util qw(:load :error &hashInvert &getSiteDirs);

use strict;
use vars qw(
    @ISA @EXPORT @EXPORT_OK %EXPORT_TAGS 
    $VIEW_PRIV $VIEWPUBL_PRIV $EDIT_PRIV $PUBL_PRIV $AUTHZ_PRIV $TMPL_PRIV 
    $ADMIN_PRIV @PRIV @APRIV @privHierarchy %privlist
);

@ISA = qw(Exporter);
@EXPORT = ();
@EXPORT_OK = qw(
    &authzMatch
    &authzGetUserHash
    &authzAdmin
    $VIEW_PRIV
    $VIEWPUBL_PRIV
    $EDIT_PRIV
    $PUBL_PRIV
    $TMPL_PRIV
    $AUTHZ_PRIV
    $ADMIN_PRIV
    @PRIV
    @APRIV
    @privHierarchy
    &authzLoad
    &authzSave
    &authzObjCheck
    &squashPriv
    &refresh_htaccess
);
%EXPORT_TAGS = ( 
    common => [qw(&authzMatch &authzGetUserHash &authzAdmin
        $VIEW_PRIV $VIEWPUBL_PRIV $EDIT_PRIV $PUBL_PRIV 
        $AUTHZ_PRIV $TMPL_PRIV $ADMIN_PRIV)],
    all =>    [@EXPORT_OK],
);

# Constants
BEGIN {
    $VIEW_PRIV = 'view';
    $VIEWPUBL_PRIV = 'viewpubl';
    $EDIT_PRIV = 'edit';
    $PUBL_PRIV = 'publish';
    $AUTHZ_PRIV = 'permissions';
    $TMPL_PRIV = 'templates';
    $ADMIN_PRIV = 'admin';

    # Collections
    @PRIV = ($VIEW_PRIV, $VIEWPUBL_PRIV, $EDIT_PRIV, $PUBL_PRIV, $AUTHZ_PRIV, $TMPL_PRIV);
    @APRIV = (@PRIV, $ADMIN_PRIV);
    # The first four privileges (view, edit, publ, authz) are hierarchical, each
    #   including all the ones before
    @privHierarchy = ($AUTHZ_PRIV, $PUBL_PRIV, $EDIT_PRIV, $VIEW_PRIV);
    %privlist = (
        $VIEW_PRIV => [ $VIEW_PRIV ],
        $EDIT_PRIV => [ $VIEW_PRIV, $EDIT_PRIV ],
        $PUBL_PRIV => [ $VIEW_PRIV, $EDIT_PRIV, $PUBL_PRIV ],
        $AUTHZ_PRIV => [ $VIEW_PRIV, $EDIT_PRIV, $PUBL_PRIV, $AUTHZ_PRIV ],
        $TMPL_PRIV => [ $TMPL_PRIV ],
        $ADMIN_PRIV => [ @APRIV ],
    );
}

#
# Return the authz hashref applicable to $fs_file, or undef on failure.
#
sub authzLoad
{
    my ($fs_file) = @_;
    my $Config = get_config();
    fileLoad($Config->{EWOK_AUTHZ_OBJECT}, $fs_file, 
        { inherit => 1, eval => 1, name => 'authz', notfound => {}, 
          addpath => undef, });
}

#
# Returns true if $user (or $req->{user}) is an admin, false otherwise.
#
sub authzAdmin
{
    my $req = shift;
    my $user = shift || $req->{user};
    my $Config = get_config();
    return $Config->{EWOK_ADMINS}->{$user} 
      if ref $Config->{EWOK_ADMINS} eq 'HASH';
    return 0;
}

#
# Return the authz value for $user from the authz file applicable
#   to $fs_file. Return 0 if $user entry does not exist, or undef
#   for any other error.
#
sub authzGetUser
{
    my $req = shift;
    my $fs_file = shift;
    my $user = shift;
    $user ||= $req->{user};
    return undef if ! $user;

    my $authz = authzLoad($fs_file);
    return undef if ref $authz ne 'HASH';
    return 0 if not exists $authz->{$user};

    return $authz->{$user};
}

#
# Return true if the authz entry for $fs_file for $user includes $priv.
#   Calls authzGetUser to get the entry in question.
#
sub authzMatch
{
    my $req = shift;
    my $fs_file = shift;
    my $priv = shift;
    my $user = shift || $req->{user};

    # Check if valid privilege
    croak "invalid privilege $priv" unless grep /^$priv$/, @APRIV;

    if ($user) {
        # Check if in EWOK_ADMINS
        my $Config = get_config();
        return 2 if ref $Config->{EWOK_ADMINS} eq 'HASH' && 
          $Config->{EWOK_ADMINS}->{$user};
    }

    # Avoid 'admin' spoofing in authz.map
    if ($priv eq $ADMIN_PRIV) {
      warning("detected attempt to spoof admin via authz.map entry for $fs_file");
      return undef;
    }

    # $VIEW_PRIV is implied by any of the others - sufficient to find anything
    $priv = '\\w+' if $priv eq $VIEW_PRIV;

    if ($user) {
        # Check the user entry
        my $userEntry = authzGetUser($req, $fs_file, $user);
        if (! defined $userEntry) {
          debug("authzGetUser returned no userEntry for $user / $fs_file");
          return undef;
        }
        my @userPriv = split /\s+/, $userEntry;
        for my $userPriv (@userPriv) {
            return 1 if grep /^$priv$/, @{$privlist{$userPriv}};
        }
        debug("no USER entry found providing $priv on $fs_file to $user");
    }

    # Check the 'ALL' entry
    my $allEntry = authzGetUser($req, $fs_file, 'ALL');
    if (! defined $allEntry) {
      debug("authzGetUser returned no userEntry for 'ALL' / $fs_file");
      return undef;
    }
    my @allPriv = split /\s+/, $allEntry;
    for my $allPriv (@allPriv) {
        return 1 if grep /^$priv$/, @{$privlist{$allPriv}};
    }
    debug("no ALL entry found providing $priv on $fs_file to $user - authzMatch failing");

    return 0;
}

#
# Return a privilege hashref for $user from the authz file applicable
#   to $fs_file, or undef on failure.
#
sub authzGetUserHash
{
    my ($req, $fs_file, $user) = @_;
    $user ||= $req->{user};
    my $authz = {};

    # Special handling for EWOK_ADMINS - give all privileges and return
    if (authzAdmin($req, $user)) {
        $authz = { map { $_ => 1 } @PRIV };
        $authz->{$ADMIN_PRIV} = 1;
        $authz->{PRIV} = $ADMIN_PRIV;
        return $authz;
    }

    # Get privs for the user and for 'ALL'
    my $userEntry = authzGetUser(@_);
    $authz->{PRIV} = $userEntry;
    my @userPriv = split /\s+/, $userEntry;
    my $allEntry = authzGetUser($req, $fs_file, 'ALL');
    $authz->{PRIV} ||= $allEntry;
    my @allPriv = split /\s+/, $allEntry;
    for my $priv (@userPriv, @allPriv) {
        if (!exists $authz->{$priv}) {
            # Add the privlist entries for this $priv to $authz
            map { $authz->{$_} = 1 } @{$privlist{$priv}};
        }
    }
    $authz->{PRIV} =~ s/$VIEWPUBL_PRIV\s*//;

    # Catch 'admin' spoofing in authz.epl
    delete $authz->{$ADMIN_PRIV} if exists $authz->{$ADMIN_PRIV};

    return $authz;
}

#
# Check the $authz privileges for all *.epl files in $file (scalar or arrayref).
#   Returns true on success (authorised) or false on failure.
#
sub authzObjCheck
{
    my $file = shift;
    my $authz = shift;
    my $op = shift;
    my $errors = shift;
    my $errCount = scalar @$errors;
    $errors = [] if ref $errors ne 'ARRAY';
    warning("Error: bad authz argument to authzObjCheck.", $errors)
        if ref $authz ne 'HASH';
    return undef if scalar @$errors > $errCount;

    my $Config = get_config();
    my @filelist = ((ref $file eq 'ARRAY') ? @$file : ( $file ));
    for my $filename (@filelist) {
        next if $filename !~ /\.epl\s*$/o;
        $filename =~ s/^\s*(.*?)\s*$/$1/o;
        # Remove leading username if paste
        $filename =~ s/^.*-([^\-]+\.epl)$/$1/o if $op eq 'paste';
        if ($Config->{EWOK_ADMIN_OBJECTS}->{$filename}) {
            info("Error: you are not authorised to \l$op $filename.", $errors)
                unless defined $authz->{$ADMIN_PRIV};
        }
        elsif ($filename eq $Config->{EWOK_AUTHZ_OBJECT}) {
            info("Error: you are not authorised to \l$op $filename.", $errors)
                unless defined $authz->{$AUTHZ_PRIV};
        }
        elsif ($filename =~ $Config->{EWOK_TEMPLATE_RE}) {
            info("Error: you are not authorised to \l$op $filename.", $errors)
                unless defined $authz->{$TMPL_PRIV};
        }
    }
    return (scalar @$errors <= $errCount);
}

#
# Check and remove implied permissions from the $authz hash
#
sub squashPriv
{
    my $authz = shift;

    # Check the authz hash for implied values
    my ($user, $privstr, @priv);
    while (($user, $privstr) = each %$authz) {
        @priv = split /\s+/, $privstr;
        next if scalar @priv <= 1;
        for (my $i = 0; $i <= $#privHierarchy-1; $i++) {
            if (grep /^$privHierarchy[$i]$/, @priv) {
                # Remove all other items in the privHierarchy
                for (my $j = $i+1; $j <= $#privHierarchy; $j++) {
                    @priv = grep !/^$privHierarchy[$j]$/, @priv;
                }
                last;
            }
        }
        $authz->{$user} = join(' ', @priv);
    }
}

#
# Save the given authz hashref into the authz file within the given directory.
#   Also defines or deletes a .htaccess file in that directory, depending
#   on whether ALL have view privilege in the authz hashref.
# Returns undef on failure, true on success.
#
sub authzSave
{
    my ($req, $dirname, $authz, $arg) = @_;
    croak "invalid directory $dirname"
        if $dirname eq '' || index($dirname,'/') < 0;
    croak "invalid permissions '$authz'"
        if ref $authz ne 'HASH';
    croak "invalid arguments '$arg'"
        if $arg && ref $arg ne 'HASH';
    my $publish = $arg->{publish} || '';
    my $Config = $arg->{config} = get_config();

    # Make sure dirname is a directory
    if (! -d $dirname) {
        $dirname =~ s|/[^/]+$||o;
    }
    else {
        $dirname = substr($dirname, 0, -1) if substr($dirname, -1) eq '/';
    }
    croak "'$dirname' is not a directory"
        unless -d $dirname;

    # Remove implied permissions from the authz hash
    squashPriv($authz);

    # If $publish, demote privileges (edit -> viewpubl, publish -> edit)
    if ($publish) {
        for my $user (keys %$authz) {
            # Map viewpubl-$publish to viewpubl and remove all viewpubl-other
            $authz->{$user} =~ s/$VIEWPUBL_PRIV-$publish/$VIEWPUBL_PRIV/;
            $authz->{$user} =~ s/$VIEWPUBL_PRIV-\w+/ /go;
            
            # Ignore ALL/NONE (GMC: is this right?!!!)
            next if $user eq 'ALL' || $user eq 'NONE';
            # Demote privileges (no 'publish' once published!)
            $authz->{$user} =~ s/\b$EDIT_PRIV\b/$VIEWPUBL_PRIV/o;
            $authz->{$user} =~ s/\b$PUBL_PRIV|$AUTHZ_PRIV\b/$EDIT_PRIV/o;
        }
    }

    # Store the authz hash
    open AUTHZ, ">", "$dirname/$Config->{EWOK_AUTHZ_OBJECT}" 
        or croak "unable to open $dirname/$Config->{EWOK_AUTHZ_OBJECT} for write: $!";
    $Data::Dumper::Indent = 1;
    print AUTHZ Data::Dumper->Dump([$authz], [qw(authz)]);
    close AUTHZ;
    chmod 0600, "$dirname/$Config->{EWOK_AUTHZ_OBJECT}";

    refresh_htaccess($dirname, $authz, $arg);
}

# Check the htaccess file against the current authz.map
sub refresh_htaccess
{
    my ($dirname, $authz, $arg) = @_;
    croak "[refresh_htaccess] invalid directory $dirname"
        if $dirname eq '' || index($dirname,'/') < 0;
    croak "[refresh_htaccess] invalid arguments '$arg'"
        if $arg && ref $arg ne 'HASH';
    my $publish = $arg->{publish} || '';
    my $Config = $arg->{config} || get_config();

    # Make sure dirname is a directory
    if (! -d $dirname) {
        $dirname =~ s|/[^/]+$||o;
    }
    else {
        $dirname = substr($dirname, 0, -1) if substr($dirname, -1) eq '/';
    }
    croak "[refresh_htaccess] '$dirname' is not a directory"
        unless -d $dirname;

    # Load authz if not passed in
    $authz ||= authzLoad($dirname);

    # Ignore only $ignore_priv in setting up .htaccess
    my $ignore_priv = $publish ? $VIEW_PRIV : $VIEWPUBL_PRIV;

    # Any ALL privilege except $ignore_priv means no .htaccess is required
    my $all = exists $authz->{ALL} && $authz->{ALL} !~ m/^\s*$ignore_priv\s*$/;

    # Update or delete .htaccess file according to $all
    if ($all) {
        croak "cannot unlink existing $dirname/.htaccess" 
            if -f "$dirname/.htaccess" && ! unlink("$dirname/.htaccess");
    }
    # Not viewable by ALL - set up .htaccess
    elsif (!  open HTACCESS, ">", "$dirname/.htaccess") {
        croak "cannot create .htaccess for $dirname: opening $dirname/.htaccess for write failed: $!";
    }
    # Opens okay, create the .htaccess
    else {
        # NONE privilege is never relevant, and we've already checked ALL
        my %users = map { $_ => 1 } grep !/^NONE|ALL$/, keys %$authz;
        # Omit any users with only $ignore_priv
        for (keys %users) {
            delete $users{$_} if $authz->{$_} eq $ignore_priv;
        }
        # Add in ewok admins
        for (keys %{$Config->{EWOK_ADMINS}}) {
            $users{$_} = 1 unless exists $users{$_};
        }
        # Save
        print HTACCESS 'require user ' . join(' ', sort keys %users) . "\n";
        close HTACCESS;
    }

    return 1;
}

#
# Ensure all stage directories get initialised with default permissions
#
BEGIN
{
    my $Config = get_config();
    if (ref $Config->{EWOK_ADMINS} eq "HASH" && keys %{$Config->{EWOK_ADMINS}}) {
        for my $dir (getSiteDirs(type => 'stage')) {
            if (-d $dir && ! -f "$dir/$Config->{EWOK_AUTHZ_OBJECT}") {
                my $default = {
                    ALL => 'viewpubl',
                    NONE => 'view',
                };
                # Default admins to permissions access (basically for view)
                $default->{$_} = 'permissions' 
                    for keys %{$Config->{EWOK_ADMINS}};
                # Save
                authzSave(undef, $dir, $default);
            }
        }
    }
}

1;

__END__

=head1 NAME

Ewok::Authz - ewok authorisation module


=head1 SYNOPSIS

    # TODO


=head1 DESCRIPTION

Ewok::Authz provides authorisation and permission handling routines
for the rest of ewok.


=head1 AUTHOR

Gavin Carr <gavin@openfusion.com.au>


=head1 COPYRIGHT

Copyright 2003, Open Fusion Pty. Ltd., Australia. All Rights Reserved.

This program is free software. You may copy or redistribute it under the same 
terms as perl itself.

=cut
