# tag: ewok configuration module (ewok/conf version)

package Ewok::Config;

use Config::Directory 0.02;

require Exporter;
use vars qw(
  @ISA @EXPORT @EXPORT_OK %EXPORT_TAGS
  $EWOK_HOME @config_read @config_write $config_multiline
);
use strict;

@ISA = qw(Exporter);
@EXPORT = qw($EWOK_HOME get_config);
@EXPORT_OK = qw(@config_read @config_write $config_multiline);
%EXPORT_TAGS = ( 
  all => [ @EXPORT, @EXPORT_OK ], 
);

@config_read = qw(
  EWOK_HOME
  EWOK_ADMINS EWOK_ADMIN_MAILTO EWOK_MAILDOMAIN
  EWOK_TOP_DIRS_HIDE EWOK_ALL_DIRS_HIDE
  EWOK_STAGE_COLOR EWOK_PUBLISH_COLOR EWOK_EWOK_COLOR
  EWOK_DELETE_INITIAL_HEADING 
  EWOK_AUTO_CONTENTS_STRING EWOK_MARKUP_DOCTYPE
  EWOK_EDIT_RE EWOK_WRAP_RE EWOK_TEMPLATE_RE EWOK_NONMETA_RE
  EWOK_ERROR_LEVEL EWOK_DEBUG_MODULES
  EWOK_TOP_DIRS_RESTRICT EWOK_TOP_DIRS_BLOCK
  APACHE_CONF APACHE_AUTH
);
@config_write = qw(
  EWOK_ADMINS EWOK_ADMIN_MAILTO EWOK_MAILDOMAIN
  EWOK_TOP_DIRS_HIDE EWOK_ALL_DIRS_HIDE
  EWOK_STAGE_COLOR EWOK_PUBLISH_COLOR EWOK_EWOK_COLOR
  EWOK_DELETE_INITIAL_HEADING 
  EWOK_AUTO_CONTENTS_STRING EWOK_MARKUP_DOCTYPE
  EWOK_EDIT_RE EWOK_WRAP_RE EWOK_TEMPLATE_RE EWOK_NONMETA_RE
  EWOK_ERROR_LEVEL EWOK_DEBUG_MODULES
  EWOK_TOP_DIRS_RESTRICT EWOK_TOP_DIRS_BLOCK
  APACHE_AUTH
);
$config_multiline = { map { $_ => 1 } qw(
  EWOK_AUTO_CONTENTS_STRING APACHE_AUTH 
  EWOK_TOP_DIRS_RESTRICT EWOK_TOP_DIRS_BLOCK
) };

my $Config = {};
my $config_ts = 0;

# Ewok base directory
$EWOK_HOME = $ENV{EWOK_HOME};
die "missing EWOK_HOME environment variable" unless $EWOK_HOME;
die "missing config directories" 
  unless -d "$EWOK_HOME/conf/dist" && -d "$EWOK_HOME/conf/local";

my @config_array = qw(EWOK_DEBUG_MODULES EWOK_ALL_DIRS_HIDE EWOK_TOP_DIRS_HIDE);
my @config_hash = qw(EWOK_ADMINS EWOK_PAGES);
my @config_map = qw(EWOK_TOP_DIRS_RESTRICT EWOK_TOP_DIRS_BLOCK);

# Refetch up-to-date config data if changed
sub get_config
{
    # This should be >, but under mod_perl -M ts's can be negative!?
    if ($config_ts != -M "$EWOK_HOME/conf/local") {
      $Config = Config::Directory->new([
          "$EWOK_HOME/conf/dist", "$EWOK_HOME/conf/local",
      ]);

      # -------------------------------------------------------------------
      # Preprocess data into more useful datatypes
      # Convert to arrayrefs
      for my $a (@config_array, @config_hash) {
        next unless $Config->{$a};
        $Config->{$a} =~ s/^\s+//;
        $Config->{$a} =~ s/\s+$//;
        if ($Config->{$a} =~ m/,/) {
          $Config->{$a} = [ split /\s*,\s*/, $Config->{$a} ];
        }
        else {
          $Config->{$a} = [ split /\s+/, $Config->{$a} ];
        }
      }

      # Convert to hashrefs
      for my $h (@config_hash) {
        next unless $Config->{$h};
        die "bad $h reference" unless ref $Config->{$h} eq 'ARRAY';
        $Config->{$h} = { map { $_ => 1 } @{$Config->{$h}} };
      }

      # Convert multiline maps into hashrefs
      my %map = ();
      for my $m (@config_map) {
        next unless $Config->{$m};
        for my $line (split /\r?\n/, $Config->{$m}) {
          next unless defined $line && substr($line,0,1) ne '#';
          $line =~ s/^\s+//;
          $line =~ s/\s+$//;
          my ($key, $value);
          # Split key from value
          if ($line =~ m/=>/) {
            ($key, $value) = split /\s*=>\s*/, $line, 2;
          }
          else {
            ($key, $value) = split /\s+/, $line, 2;
          }
          # Split value
          if ($value =~ m/,/) {
            $map{$key} = [ split /\s*,\s*/, $value ];
          }
          else {
            $map{$key} = [ split /\s+/, $value ];
          }
        }
        $Config->{$m} = \%map if keys %map;
      }

      # -------------------------------------------------------------------
      # Semi-fixed config data 

      # Metadata directory
      $Config->{EWOK_METADIR} = "$EWOK_HOME/var/meta";

      # Lock directory
      $Config->{EWOK_LOCKDIR} = "$EWOK_HOME/var/lock";

      # Directory to use as clipboard (cut, copy, paste)
      $Config->{EWOK_CLIPBOARD} = "$EWOK_HOME/var/clip";

      # Ewok (inherited) directory properties object
      $Config->{EWOK_DIRI_OBJECT} = '.dir.meta';
      $Config->{EWOK_DIRN_OBJECT} = '.dirn.meta';

      # Whether to publish authz.map files - allows ewok file manager to do 
      # security checking on publish directories, but means you need to 
      # explicitly block access to *map files
      # TODO: move this to per-target flag in site map
#     $Config->{EWOK_AUTHZ_PUBLISH} = 0;

      # Ewok authz object - deleted from file lists unless authz privilege
      $Config->{EWOK_AUTHZ_OBJECT} = 'authz.map';

      # Ewok publishing map object
      $Config->{EWOK_PUBLISH_OBJECT} = 'publish.map';

      # Ewok metadata map object
      $Config->{EWOK_METAMAP_OBJECT} = 'metadata.map';

      # Ewok metadata map object
      $Config->{EWOK_SITES_OBJECT} = 'sites.map';

      # Ewok base objects - manipulable only by ewok admin
      $Config->{EWOK_ADMIN_OBJECTS} = { map { $_ => 1 } qw(base.epl content.epl init.epl cleanup.epl util.epl), $Config->{EWOK_METAMAP_OBJECT} };

      # Ewok restricted access objects (see Ewok::File::restrictCheck)
      $Config->{EWOK_RESTRICTED_OBJECTS} = { map { $_ => 1 } 
        keys(%{$Config->{EWOK_ADMIN_OBJECTS}}), 
        $Config->{EWOK_AUTHZ_OBJECT}, 
        $Config->{EWOK_PUBLISH_OBJECT}, 
        $Config->{EWOK_SITES_OBJECT}, 
        '.htaccess',
      };

      # Locking parameters
      $Config->{EWOK_LOCKTYPE} = 'DBM';

      $config_ts = -M "$EWOK_HOME/conf/local";
    }

    return $Config;
}

1;
