# tag: ewok file operations module

package Ewok::File;

require Exporter;
use File::Basename;
use File::Copy;
use File::Find;
use File::DirSync;
use URI::Escape;
# use POSIX qw(strftime);
use Carp;
use Ewok::Config;
use Ewok::Util qw(:all);
use Ewok::Meta qw(:all);
use Ewok::Authz qw(authzObjCheck authzSave authzGetUserHash);
use Ewok::Filter qw(filter);
use Ewok::Locker;
use Ewok::Logger;

@ISA = qw(Exporter);
@EXPORT = qw();
@EXPORT_OK = qw(
    &readdir
    &relativise
    &upload
    &create
    &remove
    &rename
    &cutcopy
    &pastefiles
    &pastecheck
    &paste
    &restrictCheck
);
%EXPORT_TAGS = (
    all => [@EXPORT_OK],
    file => [qw(&upload &remove &rename &cutcopy 
        &pastecheck &paste &restrictCheck)],
);

use strict;

#
# Return an arrayref of directories and an arrayref of files comprising the
#   contents of the given directory.
#
sub readdir
{
    my ($fs_dir, $arg) = @_;
    my $web_dir = $arg->{web_dir};
    my $admin = $arg->{admin};
    my $type = $arg->{type};
    my $add_only = $arg->{add_only};
    $web_dir = $fs_dir if ! defined $web_dir;

    my (@f, @d);
    # Read files in (unless $add_only is set)
    if (! $add_only) {
        return ([], []) if ! -d $fs_dir;
        return ([], []) if ! opendir DIR, $fs_dir;
        my @files = CORE::readdir(DIR);
        closedir DIR;

        # Omit dot files if not admin
        @files = grep(!/^\./, @files) if ! $admin;
        @files = grep(!/^\.{1,2}$/, @files) if $admin;
        # Divide into directories and files
        for my $entry (sort @files) {
            if (-d "$fs_dir/$entry") {
                push @d, $entry if ! $type || $type ne 'F';
            } else {
                push @f, $entry if ! $type || $type ne 'D';
            }
        }
    }

    # Add or remove special-case directories
    my $Config = get_config();
    if (! $type || $type ne 'F') {
        @d = grep { my $foo = $_; ! grep /^\Q$foo\E$/, @{$Config->{EWOK_ALL_DIRS_HIDE}} } @d;
        if ($web_dir eq '/' || $web_dir eq '') {
            @d = grep { my $foo = $_; ! grep /^\Q$foo\E$/, @{$Config->{EWOK_TOP_DIRS_HIDE}} } @d;
            push @d, 'ewok' if $admin;
            # Add aliases at top-level
            @d = sort @d, grep !/^ewok$/, getSiteAliases();
        }
    }

    return \@d, \@f;
}

#
# Returns an arrayref of siblings to $fs_file and a hashref containing
#   their titles. Does not return $fs_file itself unless the 'self' arg
#   is set in the $arg hash, in which case the title is always 'SELF' for 
#   easy identification.
#
sub getSiblings
{
    my ($req, $fs_file, %arg) = @_;
    my $sort = $arg{'sort'};
    my $self = $arg{self};
    my $fs_dir = $fs_file;
    $fs_dir =~ s|[^/]+/*$||;
    return undef if ! -d $fs_dir;

    my @sib = ();
    my %title = ();
    my $meta = {};
    my ($d, $f) = Ewok::File::readdir($fs_dir, { type => 'F' });
    for my $file (@$f) {
        # Check for self
        if ("$fs_dir$file" eq $fs_file) {
            next if ! $self;
            push @sib, $file;
            $title{$file} = 'SELF';
        }
        else {
            $meta->{$file} = getMeta("$fs_dir$file");
            if (ref $meta eq 'HASH' && $meta->{$file}->{title}) {
                push @sib, $file;
                $title{$file} = $meta->{$file}->{title};
            }
        }
    }

    @sib = sort { 
        my $cmp = ($meta->{$a}->{item_order} || 5) cmp ($meta->{$b}->{item_order} || 5); 
        return $cmp if $cmp != 0; 
        $cmp = $meta->{$a}->{title} cmp $meta->{$b}->{title}; 
        return $cmp if $cmp != 0; 
        $a cmp $b 
    } @sib if $sort;

    return wantarray ? (\@sib, \%title) : \@sib;
}

#
# Return the absolute target $t made relative to directory $d
#   e.g. /a/b/c/d/e.html from /a/b/c/x/y/z is ../../../d/e.html
#
sub relativise ($$)
{
    my ($t, $d) = @_;
    # Return as is if $t is not absolute
    return $t if $t !~ m|^\s*/|;

    my $t1 = $t;
    $t1 =~ s|/[^/]*$|| if ! -d $t;
    my $d1 = $d;
    $d1 = substr($d1,0,-1) if substr($d1,-1) eq '/';
    my $count = 0;
    while ($d1) {
        # If d1 now matches the beginning of $t1, we're done
        last if $t1 =~ m/^$d1/;
        # Remove the final chunk
        $d1 =~ s|/[^/]+$||;
        $count++;
    }
    return undef if ! $d1;

    # Remove the matching prefix from $t
    $t =~ s|^$d1/?||;
    $t = '../' x $count . $t;
    $t = substr($t,0,-1) if substr($t,-1) eq '/';
    # Special case for current directory
    $t = '.' if $t eq '';

    return $t;
}

#
# Check for restrictions on special objects - either in EWOK_RESTRICTED_OBJECTS
#   or metadata (=~ /\.meta$/). Called by create, upload, edit, remove, 
#   rename, cutcopy, paste, and publish.
# Returns true if okay to go ahead, false otherwise, with errors placed
#   in $errors arrayref, if supplied.
#
sub restrictCheck {
    my ($req, $op, $fs_dir, $filelist, $arg) = @_;
    my $errors = $arg->{errors} || [];
    $errors = [] if ref $errors ne 'ARRAY';
    my $authz = $arg->{authz};
    my $web_dir = $arg->{web_dir};
    my $in_wd = $web_dir ? " in " . substr($web_dir,0,-1) : '';
    my $prune = $arg->{prune};
    $op ||= 'manipulate';

    # Check general privilege levels
    my $Config = get_config();
    $authz ||= authzGetUserHash($req, $fs_dir);
    if (ref $authz ne 'HASH' || 
           ($fs_dir ne $Config->{EWOK_CLIPBOARD} && ! $authz->{edit}) || 
           ($op eq 'publish' && ! $authz->{publish})) {
        info("Error: insufficient privilege to $op$in_wd.", 
            "(files: " . join(',',@$filelist) . ')', $errors);
        return 0;
    }

    # Check the $fs_dir directory is valid
    if (! ewokdir($fs_dir)) {
        my $dir = " '$web_dir'" if $web_dir;
        info("Error: cannot $op - directory$dir is not an ewok directory.",
          $errors);
        return 0;
    }

    my $okay = 1;
    my @newlist = ();
    for my $path (@$filelist) {
        # Paranoia
        $path = uri_unescape($path);
        my $file = basename($path);
        my $meta = qr/\.meta$/;
        # Fail by default on restricted objects - add specific allow cases first
        if ($Config->{EWOK_RESTRICTED_OBJECTS}->{$file} || 
                $file =~ m/$meta/ || 
                $file =~ m/$Config->{EWOK_TEMPLATE_RE}/) {
            # Add $path to @newlist
            push @newlist, $path;
            # Anyone can delete or trash metadata
            next if $file =~ /$meta/ && ($op eq 'delete' || $op eq 'trash');
            # People with 'permissions' priv can delete/cut/copy/paste authz.map
            next if ($file eq $Config->{EWOK_AUTHZ_OBJECT} || 
                     $file eq '.htaccess') && 
                 $authz->{permissions} &&
                ($op eq 'delete' || $op eq 'cut' || $op eq 'trash' || 
                 $op eq 'copy' || $op eq 'paste');
            # People with 'templates' priv can do anything to templates except publish
            next if $file =~ m/$Config->{EWOK_TEMPLATE_RE}/ && 
                 $authz->{templates} && $op ne 'publish';

            # Admins can only edit $Config->{EWOK_PUBLISH_OBJECT}
            next if $file eq $Config->{EWOK_PUBLISH_OBJECT} && 
                $authz->{admin} && $op eq 'edit';
            # Admins can do most things to EWOK_ADMIN_OBJECTS
            next if $Config->{EWOK_ADMIN_OBJECTS}->{$file} && $authz->{admin} &&
                ($op eq 'create' || $op eq 'upload' || $op eq 'edit' ||
                 $op eq 'delete' || $op eq 'cut' || $op eq 'trash' || 
                 $op eq 'copy' || $op eq 'paste');
            # Admins can cut/copy/paste/rename metadata
            next if $file =~ /$meta/ && $authz->{admin} &&
                ($op eq 'cut' || $op eq 'copy' || $op eq 'paste' || 
                 $op eq 'paste as' || $op eq 'rename');

            # Fail - note error or (quietly) prune
            $okay = 0;
            if ($prune) {
                pop @newlist;
            }
            else {
                info("Error: cannot $op restricted file $file.", $errors);
            }
        }
        else {
            push @newlist, $path;
        }
    }

    # If prune replace filelist with newlist, if different
    if ($prune) {
        $filelist = splice @$filelist, 0, $#$filelist+1, @newlist if ! $okay;
        return 1;
    }
    # Otherwise return status
    else {
        return $okay;
    }
}

#
# Actions performed after the creation of a new object (currently 
#   auto_publishing). Called by create, renameOne, upload, and paste.
#
sub postNew ($$$;$) {
    my ($req, $fs_dir, $filelist, $arg) = @_;
    $filelist = [ $filelist ] if ! ref $filelist;
    my $dirmeta = $arg->{dirmeta};
    my $web_dir = $arg->{web_dir} || $fs_dir;
    my $errors = $arg->{errors} || [];

    $dirmeta = getMeta($fs_dir, { type => 'dir', inherit => 'I'}) 
        if ! ref $dirmeta;
    return if ! $dirmeta->{auto_publish};

    # Split suffixes and remove any leading '.'
    my @list = map { (substr($_, 0, 1) eq '.') ? substr($_,1) : $_ } split /\s*[,|\s]\s*/, $dirmeta->{auto_publish};
    my $autopublish = join('|',@list);

    # Publish
    $filelist = [ $filelist ] if ref $filelist ne 'ARRAY';
    publish($req, $filelist, $fs_dir, 
        { regex => ".$autopublish\$", errors => $errors });
}

#
# Actions performed after the deletion of an object (currently 
#   auto_publishing).
# Called by delete and cutcopy.
#
sub postDelete ($$$;$) {
    my ($req, $fs_dir, $filelist, $arg) = @_;
    my $dirmeta = $arg->{dirmeta};
    my $web_dir = $arg->{web_dir} || $fs_dir;
    my $errors = $arg->{errors} || [];

    $dirmeta = getMeta($fs_dir, { type => 'dir', inherit => 'I' }) 
        if ! ref $dirmeta;
    return if ! $dirmeta->{auto_publish};

    # Split suffixes and remove any leading '.'
    my @list = map { (substr($_, 0, 1) eq '.') ? substr($_,1) : $_ } split /\s*[,|\s]\s*/, $dirmeta->{auto_publish};
    my $autopublish = join('|',@list);

    # Unpublish 
    unpublish($req, $fs_dir, $filelist, 
            { regex => "\.$autopublish\$", errors => $errors });
}

#
# Upload $infile to $fs_dir
#
sub upload ($$$;$) 
{
    my ($req, $fs_dir, $infile, $arg) = @_;
    my $web_dir = $arg->{web_dir} || $fs_dir;
    my $errors = $arg->{errors} || [];
    my $errorCount = scalar @$errors;

    my $name = $infile;
    $name =~ s|^.*[/\\]||;
#   elog($req,"[upload] $infile -> $fs_dir ($name)");

    # Check for restricted objects
    return undef 
        if ! restrictCheck($req, 'upload', $fs_dir, [ $name ], { errors => $errors });

    # Need to get a lock on the destination file
    my $Config = get_config();
    my $l = Ewok::Locker->new();
    if ($l->lock("$fs_dir$name", $req->{user}, { errors => $errors,
            altfile => "$web_dir$name", op => 'upload' })) {     
        if (open INFILE, ">", "$fs_dir$name") {
            my $buffer;
            print INFILE $buffer
                while read($infile, $buffer, 32768);
            close INFILE;
            if ($name !~ m/$Config->{EWOK_NONMETA_RE}/i) {
                # Setup default metadata
#               my $meta = getMetaCreate("$fs_dir$name");
                setMeta(undef, "$fs_dir$name", { create => 1 });
            }
            # Call postNew
#           postNew($req, $fs_dir, $name);
        }
        else {
            warning("Error: file upload failed.", $!, $errors);
        }
        $l->unlock("$fs_dir$name", $req->{user},
            { altfile => "$web_dir$name" });
    } 
    return (scalar(@$errors) > $errorCount) ? undef : 1;
}

#
# Return a (hopefully unique) filename using a system-wide id counter.
#   $style indicates the naming style to be used - vanilla id, or modified
#   with the year or year/month. A sprintf template may also be given to 
#   somewhat customise the name generated.
# Returns undef on failure, with errors in $errors arrayref, if given.
#
# TODO: this is obsolete, reliant on database. Revisit.
#
sub generateName
{
    return;

    my ($req, $style, $arg) = @_;
    my $template = $arg->{template};
    my $dir = $arg->{dir};
    my $errors = $arg->{errors};
    $errors = [] if ref $errors ne 'ARRAY';

    $style = lc($style);
    $template ||= ($style eq 'id') ? 'file-%06d.html' : '%s-%06d.html';
    my @styles = qw(id year quarter month day);
    if (!grep /^$style$/, @styles) {
        info("Error: invalid name style '$style'.", $errors);
        return undef;
    }

    # Generate our id
#   my $set;
#   *set = DBIx::Recordset->Insert({
#       '!DataSource'   => 'ewok',
#       '!Table'        => 'id',
#       '!Serial'       => 'id',
#       dir             => $dir,
#   });
#   my $id = $set->LastSerial();
#   if (! $id) {
#       warning("Error: failure generating new file id.");
#       return undef;
#   }

    # Styles - generate extra tag
#   my $tag;
#   if ($style eq 'year') {
#       $tag = strftime("%Y", localtime);
#   }
#   elsif ($style eq 'quarter') {
#       $tag = strftime("%Yq", localtime);
#       $tag .= int((strftime("%m", localtime) - 1) / 3) + 1;
#   }
#   elsif ($style eq 'month') {
#       $tag = strftime("%Y%m", localtime);
#   }
#   elsif ($style eq 'day') {
#       $tag = strftime("%Y%m%d", localtime);
#   }

    # Apply the template
#   return ($style eq 'id') ? sprintf($template, $id) : 
#                             sprintf($template, $tag, $id);
}

#
# Create an element of $type ([dfl] for dir/file/link) called $name in $fs_dir.
# Return the name of the element on success (esp. useful if generated), or 
#   undef on failure, with error messages in the $errors arrayref if given.
#
sub create
{
    my $arg = shift;
    carp "[create] argument not hashref" if ref $arg ne 'HASH';
    my $req = $arg->{req} || {};
    my $fs_dir = $arg->{fs_dir} || $arg->{dir};
    my $name_in = $arg->{name};
    my $name_style = $arg->{name_style};
    my $name_template = $arg->{name_template};
    my $type = $arg->{type} || 'f';
    my $errors = $arg->{errors};
    my $wd = $arg->{wd} || $fs_dir;
    my $target = $arg->{target};
    my $text = $arg->{text};
    my $meta = $arg->{meta};
    my $mode = $arg->{mode};
    my $markup_text = $arg->{markup_text};
    my $text_filter = $arg->{text_filter};
    my $text_filter_args = $arg->{text_filter_args};

    # Escape, trim and untaint, as required
    ($fs_dir) =~ m!\s*(.*?)/?\s*!;
    my ($name) = (uri_unescape($name_in) =~ m/\s*(.*?)\s*$/);

    # Accumulate errors locally anyway if not given arrayref
    $errors = [] if ref $errors ne 'ARRAY';
    my $errorCount = scalar @$errors;

    # Error checking
    info("Error: cannot find directory $wd.", $errors)
        if (! -d $fs_dir);
    info("Error: cannot write to directory $wd.", $errors)
        if (! @$errors && ! -w $fs_dir);
    info("Error: unknown type '$type' to create.", $errors)
        if ($type !~ m/^[dlf]$/);
    info("Error: element name cannot contain directory components.", $errors)
        if ($name =~ m|/|);
    info("Error: no name argument given to create.", $errors)
        if (! $name && ! $name_style);
    info("Error: cannot set both 'name' and 'name_style' arguments.", $errors)
        if ($name && $name_style);
    info("Error: cannot create hidden elements.", $errors)
        if ($name =~ m/^\./);
    info("Error: no target given for link.", $errors)
        if ($type eq 'l' && ! $target);
    info("Error: cannot give text when creating " . 
            (($type eq 'l') ? 'link' : 'directory') . '.', $errors)
        if ($text && ($type eq 'l' || $type eq 'd'));
    restrictCheck($req, 'create', $fs_dir, [ $name ], { errors => $errors });
    return undef if scalar(@$errors) > $errorCount;

    my $dirmeta = getMeta($fs_dir, { type => 'dir', inherit => 'I' });
    $name = generateName($req, $name_style, { errors => $errors, 
        template => $name_template, dir => $fs_dir }) if $name_style;
    info("Error: $wd/$name already exists.", $errors)
        if ! @$errors && -e "$fs_dir/$name";
    return undef if scalar(@$errors) > $errorCount;

    # File create
    my $Config = get_config();
    if ($type eq 'f') {
        my $fs_file = "$fs_dir/$name";
        if (! open FILE, ">>", $fs_file) {
            warning("Error: unable to create $wd/$name.", $errors);
            return undef;
        }
        else {
            if ($text) {
                $text = filter($text) if $markup_text;
                print FILE $text;
            }
            close FILE;
            chmod $mode, $fs_file if $mode;
            if ($fs_file !~ m/$Config->{EWOK_NONMETA_RE}/i) {
                # Setup default metadata
                setMeta($meta, $fs_file, { create => 1, errors => $errors });
            }
        }
    }

    # Symlink create
    elsif ($type eq 'l') {
        my $wt = $target;
        # Map (web-relative) absolute target to filesystem
        if ($target =~ m|^\s*/|) {
            if (ref $req->{req_rec}) {
                $target = $req->{req_rec}->lookup_uri($target)->filename;
            }
            else {
                info("Error: unable to map $wt - please try a relative link.",
                  $errors);
                return undef;
            }
        }
        # Or map relative target to filesystem
        else {
            if ($wd ne $fs_dir && ref $req->{req_rec}) {
                $target = "$wd/$target";
                $target = $req->{req_rec}->lookup_uri($target)->filename;
            }
            # If we can't do a lookup try and fake it - this could be dangerous
            else {
                $target = "$fs_dir/$target";
            }
        }

        # Now turn the absolute fs target into a relative one
        $target = relativise($target, $fs_dir);

        # Check target permissions
        my $authz = authzGetUserHash($req, "$fs_dir/$target");
        if (! $authz->{view}) {
            info("Error: cannot create link to $wt - you do not have view permissions.", $errors);
            return undef;
        }
        # Check target exists
        elsif (! -e "$fs_dir/$target") {
            info("Error: cannot create link to $wt - does not exist.", $errors);
            return undef;
        }
        # Check target isn't a link
        elsif (-l "$fs_dir/$target") {
            info("Error: cannot create link to $wt - it's a link.", $errors);
            return undef;
        }
        # Make symlink
        elsif (! symlink($target, "$fs_dir/$name")) {
            warning("Error creating link to $wt.", $!, $errors);
            return undef;
        }
        elsif ("$fs_dir/$name" !~ m/$Config->{EWOK_NONMETA_RE}/i) {
            # Setup default metadata
            setMeta($meta, "$fs_dir/$name", { create => 1, errors => $errors });
        }
    }
     
    # Directory create
    elsif ($type eq 'd') {
        if (! mkdir("$fs_dir/$name", 0750)) {
            warning("Error on directory create.", $!, $errors);
            return undef;
        }

        # Setup default metadata
        setMeta($meta, "$fs_dir/$name", { create => 1, errors => $errors });
    }

#   postNew($req, $fs_dir, $name, { dirmeta => $dirmeta });

    return $name;
}

#
# Callback for directory unlinks via File::Find
#
sub remove_unlink
{
    my $obj = $File::Find::name;
    $obj ||= shift;
    ($obj) = ($obj =~ m|^\s*(.*?)\s*$|);
    if (-d $obj) {
        rmdir $obj;
    }
    elsif (-e $obj) {
        unlink $obj;
    }
    else {
        croak("[remove_unlink] '$obj' not found");
    }
}

#
# Remove the given elements in $fs_dir: the directories in @$dirlist, and the
#   files in @$filelist. Directory has to be empty to be deleted (although
#   we delete dotfiles automatically).
# Returns true on success or undef on failure, with errors returned in the
#   @$errors arrayref (if provided).
#
sub remove
{
    my $req = shift;
    my $fs_dir = shift;
    my $filelist = shift;
    my $arg = shift;
    my $errors = $arg->{errors};
    my $web_dir = $arg->{web_dir};
    my $unlink = $arg->{unlink};
    $filelist = [] if ! $filelist;
    $filelist = [ $filelist ] if ref $filelist ne 'ARRAY';
    # Accumulate errors locally anyway if not given arrayref
    $errors = [] if ref $errors ne 'ARRAY';
    $web_dir ||= $fs_dir;
    $fs_dir  = substr($fs_dir, 0, -1)  if substr($fs_dir, -1) eq '/';
    $web_dir = substr($web_dir, 0, -1) if substr($web_dir, -1) eq '/';
    my $errorCount = scalar @$errors;

    # Error checking
    info("Error: cannot find directory '$web_dir'.", $errors)
        if (! -d $fs_dir);
    info("Error: cannot write to directory '$web_dir'.", $errors)
        if (! @$errors && ! -w $fs_dir);
    restrictCheck($req, 'delete', $fs_dir, $filelist, { errors => $errors });
    return undef if scalar(@$errors) > $errorCount;

    # Move to Trash (on the clipboard) unless $unlink is set
    if (!$unlink) {
        my $filelist2 = [ @$filelist ];
        if (@$filelist2) {
            # Cut $filelist2 to a Trash area
            cutcopy($req, $fs_dir, $filelist2, { 
                tag => "Trash-" . $req->{user}, 
                op => 'trash', 
                errors => $errors, 
                web_dir => $web_dir, 
                keep => 1
            });
        }
        # Do any postDelete stuff (original $filelist)
#       postDelete($req, $fs_dir, $filelist, { errors => $errors });
    }

    # Explicit unlink
    else {
        for my $name (@$filelist) {
            # Paranoia check
            if ($name =~ m|/|) {
                warning("Error: '$name' is not a basename ('/' is not allowed).", $errors)
            }
            else {
                # Untaint $name
                ($name) = ($name =~ m|([^/]+)$|);
                # For directories, do a recursive delete, no locking
                if (-d "$fs_dir/$name") {
                    $ENV{PATH} = '/usr/bin:/bin';
                    find({
                        wanted => \&remove_unlink, 
                        no_chdir => 1,
                        untaint => 1,
                        bydepth => 1,
                    }, "$fs_dir/$name");
                }
                else {
                    # For files, need to worry about locking
                    my $l = Ewok::Locker->new();
                    if ($l->lock("$fs_dir/$name", $req->{user},
                            { errors => $errors, altfile => "$web_dir/$name",
                              op => 'delete' })) {
                        if (! unlink "$fs_dir/$name") {
                            warning("Error: unable to delete file '$web_dir/$name'.", $errors);
                        }
                        else {
                            # Remove metadata
                            delMeta("$fs_dir/$name");
                        }
                        $l->unlock("$fs_dir/$name", $req->{user},
                            { altfile => "$web_dir/$name" });
                    }
                }
            }
        }
#       postDelete($req, $fs_dir, $filelist, { errors => $errors })
#           if ! @$errors;
    }

    return undef if @$errors;
    return 1;
}

#
# Rename (within $web_dir) the set of files in @$namesFrom to @$namesTo (uses
#   &renameOne).
# Returns true if all renames are successful, undef otherwise, with errors
#   in the @$errors arrayref (if provided).
#
sub rename ($$$$;$$)
{
    my $req = shift;
    my $namesFrom = shift;
    my $namesTo = shift;
    my $fs_dir = shift;
    my $errors = shift;
    my $web_dir = shift;
    $web_dir ||= $fs_dir;
    # Accumulate errors locally anyway if not given arrayref
    $errors = [] if ref $errors ne 'ARRAY';
    my $errorCount = scalar @$errors;

    # Error checking
    info("Error: cannot find directory '$web_dir'.", $errors)
        if (! -d $fs_dir);
    info("Error: cannot write to directory '$web_dir'.", $errors)
        if (! @$errors && ! -w $fs_dir);
    info("Error: no objects given to rename.", $errors)
        if $#$namesFrom < 0;
    info("Error: rename from and to lists of different sizes.", $errors)
        if $#$namesFrom != $#$namesTo;
    restrictCheck($req, 'rename', $fs_dir, [ @$namesFrom, @$namesTo ], 
        { errors => $errors });
    return undef if scalar(@$errors) > $errorCount;

    my %namesFrom = map { $_ => 1 } @$namesFrom;
    my (@nf, @nt);
    my $status = 1;
    my $last = $#$namesFrom;
    for (my $i = 0; $i <= $last; $i++) {
        # Use temporary files for targets that are also to be renamed
        if ($namesFrom{$namesTo->[$i]}) {
            push @$namesTo, $namesTo->[$i];
            $namesTo->[$i] .= '.tmp.' . $$;
            push @$namesFrom, $namesTo->[$i];
        }
        if (! renameOne($req, $namesFrom->[$i], $namesTo->[$i], $fs_dir, 
                $errors, $web_dir)) {
            $status = undef;
        }
    }
    # Second pass
    for (my $i = $last+1; $i <= $#$namesFrom; $i++) {
        if (! renameOne($req, $namesFrom->[$i], $namesTo->[$i], $fs_dir, 
                $errors, $web_dir)) {
            $status = undef;
        }
    }

    # postNew & postDelete
#   postNew($req, $fs_dir, $namesTo);
#   postDelete($req, $fs_dir, $namesFrom);

    return $status;
}

#
# Rename a single file/directory in $fs_dir from $nameFrom to $nameTo.
#   Returns true on success, undef on failure, with errors in the @$errors
#   arrayref (if provided).
#
sub renameOne ($$$;$$$)
{
    my $req = shift;
    my $nameFrom = shift;
    my $nameTo = shift;
    my $fs_dir = shift;
    my $errors = shift;
    my $web_dir = shift;
    $web_dir ||= $fs_dir;
    $fs_dir  = substr($fs_dir, 0, -1) if substr($fs_dir, -1) eq '/';
    $web_dir = substr($web_dir, 0, -1) if substr($web_dir, -1) eq '/';
    # Accumulate errors locally anyway if not given arrayref
    $errors = [] if ref $errors ne 'ARRAY';
    my $errorCount = scalar @$errors;

    # Error checking
    info("Error: cannot find object '$web_dir/$nameFrom'.", $errors)
        if (! -e "$fs_dir/$nameFrom");
    info("Error: source filename '$nameFrom' contains invalid directory separator '/'.", $errors) if (uri_unescape($nameTo,'/') =~ m|/|);
    info("Error: target filename '$nameTo' contains invalid directory separator '/'.", $errors) if (uri_unescape($nameTo,'/') =~ m|/|);
    return undef if scalar(@$errors) > $errorCount;

    # Trim and untaint names
    $nameFrom =~ s|\s*(.*?)\s*$|$1|;
    ($nameFrom) = (uri_unescape($nameFrom) =~ m|.*?([^/]+)$|);
    $nameTo =~ s|\s*(.*?)\s*$|$1|;
    ($nameTo) = (uri_unescape($nameTo) =~ m|.*?([^/]+)$|);

    # Lock, rename, unlock
    my $l = Ewok::Locker->new();
    if ($l->lock("$fs_dir/$nameFrom", $req->{user},
                    { errors => $errors, altfile => "$web_dir/$nameFrom",
                      op => 'rename' })) {
        if ($l->lock("$fs_dir/$nameTo", $req->{user},
                    { errors => $errors, altfile => "$web_dir/$nameTo",
                      op => 'rename'  })) {
            if (!CORE::rename("$fs_dir/$nameFrom","$fs_dir/$nameTo")) {
                warning("Error: renaming '$nameFrom' to '$nameTo' in $web_dir " .
                  "failed.", $!, $errors);
            }
            # Move metadata for non-directories: get, set, delete
            elsif (! -d "$fs_dir/$nameFrom") {
                mvMeta($fs_dir, $nameFrom, $nameTo);
#               my $meta = getMeta("$fs_dir/$nameFrom");
#               if ($meta) {
#                   setMeta($meta, "$fs_dir/$nameTo");
#                   delMeta("$fs_dir/$nameFrom");
#               }
            }
            $l->unlock("$fs_dir/$nameTo", $req->{user},
                    { altfile => "$web_dir/$nameTo" });
        }
        $l->unlock("$fs_dir/$nameFrom", $req->{user},
                { altfile => "$web_dir/$nameFrom" });
    }

    return ((@$errors) ? undef : 1);
}

#
# Copy or move the files in @$filelist (in $fs_dir) to the EWOK_CLIPBOARD
#   directory. Removes any files previously on the clipboard for this user
#   (identified with $tag) i.e. only one set of files is stored for a user.
# Returns true if all copies succeeded, undef on errors, with errors in the
#   errors arrayref (if supplied).
# Filenames on the clipboard are munged simply by prefixing '$tag-'.
#
sub cutcopy
{
    my $req = shift;
    my $fs_dir = shift;
    my $filelist = shift;
    my $arg = shift;
    my $tag = $arg->{tag};
    my $op = $arg->{op};
    my $errors = $arg->{errors};
    my $web_dir = $arg->{web_dir};
    my $keep = $arg->{keep};
    my $Config = get_config();
    my $clip_dir = $Config->{EWOK_CLIPBOARD};
    $web_dir ||= $fs_dir;
    $op ||= 'copy';
    $fs_dir   = substr($fs_dir, 0, -1) if substr($fs_dir, -1) eq '/';
    $web_dir  = substr($web_dir, 0, -1) if substr($web_dir, -1) eq '/';
    $clip_dir = substr($clip_dir, 0, -1) if substr($clip_dir, -1) eq '/';
    # Accumulate errors locally anyway if not given arrayref
    $errors = [] if ref $errors ne 'ARRAY';
    my $errorCount = scalar @$errors;

    # Check error conditions
    warning("Error: clipboard directory '$clip_dir' does not exist (or is not writeable).", $errors)
        if ! -d $clip_dir || ! -w $clip_dir;
    info("Error: unknown operation '$op' ('copy', 'cut', and 'trash' supported).", $errors)
        if $op ne 'copy' && $op ne 'cut' && $op ne 'trash';
    warning("Error: required tag not set for $op.", $errors)
        if scalar(@$errors <= $errorCount) && ! $tag;
    info("Error: no files selected to $op.", $errors)
        if scalar(@$errors <= $errorCount) && ! @$filelist;
    restrictCheck($req, $op, $fs_dir, $filelist, { errors => $errors });
    return undef if scalar(@$errors) > $errorCount;

    # Remove any existing files (and metadata) on the clipboard for this tag
    chdir ($clip_dir);
    if (! $keep) {
        my @delete = <$tag-*>;
        remove($req, $clip_dir, \@delete, {
            errors => $errors, 
            web_dir => '/clip', 
            unlink => 1,
        });
        return undef if scalar(@$errors) > $errorCount;
    }

    # Copy/move each file
    for my $file (@$filelist) {
        # Trim and untaint $file
        $file =~ s|\s*(.*?)\s*$|$1|;
        ($file) = (uri_unescape($file) =~ m|.*?([^/]+)$|);
        if (-f "$fs_dir/$file" || -l "$fs_dir/$file") {
            # Copy
            my $l = Ewok::Locker->new();
            if ($op eq 'copy') {
                if (! copy("$fs_dir/$file", "$clip_dir/$tag-$file")) {
                    warning("Error: $op failed.", $!, $errors);
                }
                else {
                    my $meta = getMeta("$fs_dir/$file");
                    setMeta($meta, "$clip_dir/$tag-$file") if $meta;
                }
            }
            # Cut or trash
            elsif ($l->lock("$fs_dir/$file", $req->{user},
                    { errors => $errors, altfile => "$web_dir/$file",
                      op => 'cut' })) {
                if (! move("$fs_dir/$file", "$clip_dir/$tag-$file")) {
                    warning("Error: $op failed.", $!, $errors);
                }
                else {
                    my $meta = getMeta("$fs_dir/$file");
                    delMeta("$fs_dir/$file");
                    delMeta("$clip_dir/$tag-$file") if $keep;
                    setMeta($meta, "$clip_dir/$tag-$file") if $meta;
                }
                $l->unlock("$fs_dir/$file", $req->{user},
                    { altfile => "$web_dir/$file" });
            }
        }
        # For directories, metadata is automatic and locking is prohibitive
        elsif (-d "$fs_dir/$file") {
            # Copy
            if ($op eq 'copy') {
                my $ds = File::DirSync->new({ nocache => 1 });
                $ds->dirsync("$fs_dir/$file", "$clip_dir/$tag-$file");
            }
            # Cut or trash
            else {
                if (-d "$clip_dir/$tag-$file") {
                    remove($req, $clip_dir, "$tag-$file", {
                        errors => $errors,
                        web_dir => '/clip',
                        unlink => 1,
                    });
                }
                if (! move("$fs_dir/$file", "$clip_dir/$tag-$file")) {
                    warning("Error: $op failed.", $!, $errors);
                }
            }
        }
        else {
            info("Error: file '$web_dir/$file' does not exist.", $errors);
        }
    }
#   postDelete($req, $fs_dir, $filelist, { errors => $errors }) 
#       if $op ne 'copy';

    return ! scalar(@$errors) > $errorCount;
}

#
# Returns an arrayref of files available on the clipboard for this $tag (user),
#   or undef on error.
#
sub pastefiles ($;$)
{
    my $tag = shift;
    my $munge = shift;
    my $Config = get_config();
    my $clip_dir = $Config->{EWOK_CLIPBOARD};
    $clip_dir .= '/' if substr($clip_dir, length($clip_dir)-1, 1) ne '/';

    # Return undef if any errors
    return undef if ! -d $clip_dir || ! -r $clip_dir;
    return undef if ! $tag;

    # Glob $tag files on the clipboard
    chdir($clip_dir);
    my @filelist = sort <$tag-*>;
    map { s/^$tag-// } @filelist if $munge;
    return \@filelist;
}

#
# Checks whether there are any files to paste for this $tag (user).
# Returns the file count (0 if none), or undef on error.
#
sub pastecheck ($)
{
    my $tag = shift;
    my $filelist = pastefiles($tag);
    return (defined $filelist) ? scalar @$filelist : undef;
}

#
# Paste (copy) all $tag files to $fs_dir from the EWOK_CLIPBOARD directory.
# Returns true if all copies are successful, undef otherwise, with errors
#   returned via the $errors arrayref (if provided).
# Filenames on the clipboard are munged simply by prefixing '$tag-'.
#
sub paste
{
    my $req = shift;
    my $fs_dir = shift;
    my $arg = shift;
    my $tag = $arg->{tag} || $req->{user};
    my $authz = $arg->{authz};
    my $errors = $arg->{errors};
    my $web_dir = $arg->{web_dir};
    my $fromlist = $arg->{fromlist};
    my $tolist = $arg->{tolist};
    my $clobber = $arg->{clobber};
    my $Config = get_config();
    my $clip_dir = $Config->{EWOK_CLIPBOARD};
    $web_dir ||= $fs_dir;
    $fs_dir   = substr($fs_dir, 0, -1) if substr($fs_dir, -1) eq '/';
    $web_dir  = substr($web_dir, 0, -1) if substr($web_dir, -1) eq '/';
    $clip_dir = substr($clip_dir, 0, -1) if substr($clip_dir, -1) eq '/';
    # Accumulate errors locally anyway if not given arrayref
    $errors = [] if ref $errors ne 'ARRAY';
    my $errorCount = scalar @$errors;
    my %rename = ();

    # Check error conditions
    warning("Error: clipboard directory '$clip_dir does not exist.", $errors)
        if ! -d $clip_dir || ! -r $clip_dir;
    info("Error: cannot write to paste directory '$web_dir'.", $errors)
        if ! -w $fs_dir;
    info("Error: required paste tag not set.", $errors)
        if ! $tag;
    return undef if scalar(@$errors) > $errorCount;

    # Glob $tag files on the clipboard
    # We treat no files to paste as an error (use &pastecheck first)
    chdir($clip_dir);
    my @filelist = @{pastefiles($tag)};
    info("Error: no files found to paste for '$tag'.", $errors)
        if ! @filelist;
    if (@$fromlist && @$tolist) {
        if (scalar @filelist != scalar @$fromlist ||
            scalar @filelist != scalar @$tolist) {
            warning("Error: found " . scalar(@filelist) . 
              "files to paste but rename lists contain " . scalar(@$fromlist) .
              " and " . scalar(@$tolist) . "respectively.", $errors);
        }
        else {
            for (my $i = 0; $i <= $#$fromlist; $i++) {
                $rename{"$tag-$fromlist->[$i]"} = $tolist->[$i];
            }
        }
    }
    return undef if ! authzObjCheck(\@filelist, $authz, 'paste', $errors);
    restrictCheck($req, 'paste', $fs_dir, \@filelist, { errors => $errors });
    restrictCheck($req, 'paste as', $fs_dir, $tolist, { errors => $errors });
    return undef if scalar(@$errors) > $errorCount;

    # Paste each file
    for (my $i = 0; $i <= $#filelist; $i++) {
        # Untaint
        my ($filename) = ($filelist[$i] =~ m|([^/]+)$|);
        my $shortname = $filename;
        $shortname =~ s/^$tag-//;
        my $newname;
        if (!@$fromlist) {
            $newname = $shortname;
        }
        else {
            # Untaint
            ($newname) = (($rename{$filename} || $shortname) =~ m|([^/]+)$|);
        }
        if (-e "$clip_dir/$filename") {
            # If the target file exists, make a copy or overwrite
            if (-e "$fs_dir/$newname") {
                # Check whether to overwrite this file
                if (! $clobber->{$shortname}) {
                    my $newname2 = "Copy-Of-$newname";
                    if (-e "$fs_dir/$newname2") {
                        for my $i (2 .. 99) {
                            $newname2 = "Copy${i}-Of-$newname";
                            last if ! -e "$fs_dir/$newname2";
                        }
                        if (-e "$fs_dir/$newname2") {
                            warning("Error: paste failed (too many copies already).", $errors);
                            next;
                        }
                    }
                    $newname = $newname2;
                }
            }
            my $fs_file = "$fs_dir/$newname";
            # For directories, metadata is automatic, don't worry about locking
            if (-d "$clip_dir/$filename") {
                my $ds = File::DirSync->new({ nocache => 1 });
                $ds->dirsync("$clip_dir/$filename", $fs_file);
                # Call postNew for all files
#               $ENV{PATH} = '/usr/bin:/bin';
#               find({ 
#                   wanted => sub {
#                       postNew($req, $File::Find::dir, basename($_)) 
#                           if ! -d $File::Find::name;
#                   },
#                   no_chdir => 1,
#                   untaint => 1,
#                   bydepth => 1,
#               }, $fs_file);
            }
            else {
            # Lock the file we're pasting to
                my $l = Ewok::Locker->new();
                if ($l->lock($fs_file, $req->{user},
                        { errors => $errors, altfile => "$web_dir/$newname",
                          op => 'paste' })) {
                    if (! &copy("$clip_dir/$filename", "$fs_dir/$newname")) {
                        warning("Error: paste failed.", $!, $errors);
                    }
                    else {
                        if ($fs_file !~ m/$Config->{EWOK_NONMETA_RE}/i) {
                            # Copy metadata
                            my $meta = getMeta("$clip_dir/$filename");
                            if (defined $meta) {
                                setMeta($meta, $fs_file);
                            }
                        }
                        # Call #ostNew
#                       postNew($req, $fs_dir, $newname);
                    }

                    # Unlock the pasted file
                    $l->unlock("$fs_dir/$newname", $req->{user},
                        { altfile => "$web_dir/$newname" });
                }
            }
        }
        else {
            warning("Error: clipboard file '$newname' not found.", $errors);
        }
    }
    return ((scalar(@$errors) > $errorCount) ? undef : 1);
}

1;

__END__

=head1 NAME

Ewok::File - ewok file operations


=head1 SYNOPSIS

    # TODO


=head1 DESCRIPTION

Ewok::File provides file manipulation routines for the ewok file manager.


=head1 AUTHOR

Gavin Carr <gavin@openfusion.com.au>


=head1 COPYRIGHT

Copyright 2003, Open Fusion Pty. Ltd., Australia. All Rights Reserved.

This program is free software. You may copy or redistribute it under the same 
terms as perl itself.

=cut
