# tag: generic filter and pipeline framework

package Ewok::Filter;

use Carp;
use Ewok::Util qw(:error);
require Exporter;

@ISA = qw(Exporter);
@EXPORT = qw();
@EXPORT_OK = qw(
    filter
    register_filter
    register_pipeline
);
%EXPORT_TAGS = (
    all => [ @EXPORT_OK ],
    register => [ qw(register_filter register_pipeline) ],
);
use strict;

use vars qw(%FILTER %PIPELINE);

# %FILTER maps symbolic filter names (lowercase, by convention) to 
# subroutine references. Usually added via register_filter().
%FILTER = ();

# %PIPELINE maps symbolic names (beginning with a capital letter,
# by convention) to array references of filter names. Usually added
# via register_pipeline().
%PIPELINE = ( 
    NULL => [],
);

# Register a filter (called by filter/pipeline implementation)
sub register_filter 
{
    croak "[register_filter] odd number of parameters"
        if @_ % 2;
    while (@_) {
        my $name = shift;
        my $filter = shift;
        croak "[register_filter] $name filter $filter not subref"
            if ref $filter ne 'CODE';
        $FILTER{$name} = $filter;
    }
}

# Register a pipeline (called by filter/pipeline implementation)
sub register_pipeline
{
    croak "[register_pipeline] odd number of parameters"
        if @_ % 2;
    while (@_) {
        my $name = shift;
        my $pipeline = shift;
        croak "[register_pipeline] $name pipeline $pipeline not array ref"
            if ref $pipeline ne 'ARRAY';
        $PIPELINE{$name} = $pipeline;
    }
}

# Applies a given filter pipeline to the given content.
sub filter
{
    my ($content, $name, $arg) = @_;
    $name ||= ($content =~ m/^[^<]*<(html|!doctype|\?xml)\b/i) ?
         'StandardHTML' : 'StandardTextHTML';
    my $pipeline = $PIPELINE{$name};
    croak "invalid pipeline $name" unless ref $pipeline eq 'ARRAY';

    # Apply filters
    for my $filter (@$pipeline) {
        croak "invalid filter $filter" unless ref $FILTER{$filter} eq 'CODE';
        $content = $FILTER{$filter}->($content, $arg);
    }

    return $content;
}

1;

__END__

=head1 NAME

Ewok::Filter - generic filter and pipeline framework


=head1 SYNOPSIS

    # For content filtering:
    use Ewok::Filter qw(filter);
    $out = filter($content, $pipename, { @options });

    # Registering, by content filters:
    use Ewok::Filter qw(:register);
    register_filter(
        filter1 => \&filter1,
        filter2 => \&filter2,
    );
    register_pipeline(
        MyFilter => [ qw(filter1 filter2) ],
    );


=head1 DESCRIPTION

Ewok::Filter provides the generic filter / pipeline framework that
ewok uses to munge content for presentation or editing, etc.


=head1 AUTHOR

Gavin Carr <gavin@openfusion.com.au>


=head1 COPYRIGHT

Copyright 2003, Open Fusion Pty. Ltd., Australia. All Rights Reserved.

This program is free software. You may copy or redistribute it under the same 
terms as perl itself.

=cut
