# tag: simple directory-cluster-style view-filter

package Ewok::Filter::ClusterIndex;

use Ewok::Filter qw(:register);
use strict;

# Register filters and pipelines with Ewok::Filter
BEGIN {
    register_filter(
        cluster_index => \&filter,
    );
    register_pipeline(
        ClusterIndex => [qw(
            cluster_index
            text_headings text_bold_italics text_links
            html_named_anchors html_headings
        )], 
    );
}

# Markup the given clusters
sub markup 
{
    my $clusters = shift;
    my $out = '';

    if (@$clusters) {
        # Wrap in a table
        $out .= qq(<table border="0" align="center" cellpadding="4" cellspacing="3" width="90%"><tr>\n);
        $out .= qq(<td valign="top" width="50%">\n);

        # Markup each cluster
        for (my $i = 0; $i <= $#$clusters; $i++) {
            my $current = $clusters->[$i];
            # Cluster head
            $out .= qq(<span class="cluster_head">$current->[0]</span><br />\n);
            $out .= qq(<span class="cluster_item">\n);
            # Cluster items
            if ($#$current > 0) {
                $out .= join qq( |\n), @$current[ 1 .. $#$current ];
            }
            $out .= qq(</span><br /><br />\n);
            $out .= qq(</td><td valign="top" width="50%">\n)
                if $i == int($#$clusters / 2);
        }

        # Close table
        $out .= qq(</td>\n);
        $out .= qq(</tr></table><br />\n\n);
    }

    return $out;
}

# Markup suitable lines in the given page to a portal-cluster-index style
sub filter
{
    my $page = shift;
    my $arg = shift;

    # Parse the page
    my $prelude = '';
    my @clusters = ();
    my @current = ();
    my $postlude = '';
    for (split /\r?\n/, $page) {
        # Comments allowed
        next if m/^#/;
        # Once postlude has started it gets everything else
        if ($postlude) {
            $postlude .= "$_\n";
        }
        # New cluster
        elsif (m/^\[\s*(.*?)\s*\]\s*$/) {
            push @clusters, [ @current ] if @current;
            @current = ( $1 );
        }
        # New entry
        elsif (@current && m/^(\S.*?)\s*$/) {
            push @current, $1 if substr($1,0,1) ne '#';
        }
        # Blank line
        elsif (m/^\s*$/) {
            # Blank lines terminate clusters as long as at least one entry
            if (@current > 1) {
                push @clusters, [ @current ];
                @current = ();
            }
        }
        # Anything else
        else {
            # Add to $prelude before initial cluster
            if (! @clusters) {
                $prelude .= "$_\n";
            }
            # Anything else after a cluster terminates the parse
            else {
                $postlude .= "$_\n";
            }
        }
    }
    # Check for hanging cluster
    push @clusters, \@current if @current;

    # Markup clusters and return
    return $prelude . markup(\@clusters) . $postlude;
}

1;

__END__

=head1 NAME

Ewok::Filter::ClusterIndex - simple portal-cluster-index view-filter


=head1 SYNOPSIS

    use Ewok::Filter::ClusterIndex;

    $page = Ewok::Filter::ClusterIndex::filter($page);



=head1 DESCRIPTION

Ewok::Filter::ClusterIndex is a simple portal-cluster-index-style 
view-filter. It expects to find lines of the form:

[ link1 ]

link2
link3
link4

where the various links are typically valid HTML anchor tags (or just a 
simple text label). This is converted to:

  link1
  link2 | link3 | link4



=head1 AUTHOR

Gavin Carr <gavin@openfusion.com.au>


=head1 COPYRIGHT

Copyright 2001-2003, Open Fusion Pty. Ltd., Australia. All Rights Reserved.

This program is free software. You may copy or redistribute it under the same 
terms as perl itself.

=cut
