# tag: simple Wiki-like view-filter features

package Ewok::Filter::Wiki;

use Ewok::Filter qw(:register);
use strict;

# Register filters and pipelines with Ewok::Filter
BEGIN {
    register_filter(
        yahoo => \&filter,
    );
    register_pipeline(
        Yahoo => [qw(
            yahoo
            text_headings text_bold_italics text_links
            html_named_anchors html_headings
        )], 
    );
}

# Markup wiki-style links
sub filter
{
    my $page = shift;
    my $arg = shift;

    # Parse the page
    my $prelude = '';
    my @clusters = ();
    my @current = ();
    my $postlude = '';
    for (split /\r?\n/, $page) {
        # Once postlude has started it gets everything else
        if ($postlude) {
            $postlude .= "$_\n";
        }
        # New cluster
        elsif (m/^\[\s*(.*?)\s*\]\s*$/) {
            push @clusters, [ @current ] if @current;
            @current = ( $1 );
        }
        # New entry
        elsif (@current && m/^(\S.*?)\s*$/) {
            push @current, $1 if substr($1,0,1) ne '#';
        }
        # Blank line
        elsif (m/^\s*$/) {
            # Blank lines terminate clusters as long as at least one entry
            if (@current > 1) {
                push @clusters, [ @current ];
                @current = ();
            }
        }
        # Anything else
        else {
            # Add to $prelude before initial cluster
            if (! @clusters) {
                $prelude .= "$_\n";
            }
            # Anything else after a cluster terminates the parse
            else {
                $postlude .= "$_\n";
            }
        }
    }
    # Check for hanging cluster
    push @clusters, \@current if @current;

    # Markup clusters and return
    return $prelude . markup(\@clusters) . $postlude;
}

1;

__END__

=head1 NAME

Ewok::Filter::Wiki - simple Wiki-like view filter features


=head1 SYNOPSIS

    use Ewok::Filter::Wiki;

    $page = Ewok::Filter::Wiki::filter($page);



=head1 DESCRIPTION

Ewok::Filter::Yahoo is a simple Yahoo-menu-style view-filter. It expects
to find lines of the form:

[ link1 ]

link2
link3
link4

where the various links are typically valid HTML anchor tags (or just a 
simple text label). This is converted to:

  link1
  link2 | link3 | link4



=head1 AUTHOR

Gavin Carr <gavin@openfusion.com.au>


=head1 COPYRIGHT

Copyright 2001-2003, Gavin Carr and Open Fusion Pty. Ltd. All Rights Reserved.

This program is free software. You may copy or redistribute it under the same 
terms as perl itself.

=cut
