# tag: ewok publishing operations

package Ewok::Publish;

require Exporter;
use File::Compare;
use File::Copy;
use URI::Escape;
use Ewok::Config;
use Ewok::Util qw(:all);
use Ewok::Meta qw(getMeta);
use Ewok::Authz qw(authzObjCheck authzSave authzLoad authzGetUserHash);
use Ewok::File qw(restrictCheck);
# use Ewok::Logger;

@ISA = qw(Exporter);
@EXPORT = qw();
@EXPORT_OK = qw(
    &targetRewrite
    &getPublishDirs
    &getSiteHash
    &getPublishHash
    &publish
    &publishAll
    &restrictCheck
);
%EXPORT_TAGS = (
    all => [@EXPORT_OK],
);

use strict;

#
# Rewrite $web_file using the site map, but checking it against the
#   'target' directory rather than the 'source' one. This is useful for 
#   viewing remote target directory content.
#
sub targetRewrite
{
    my $web_file = shift;

    # Load site map
    my $smap = loadSiteMap();
    return $web_file unless ref $smap eq 'HASH';

    # Iterate over sites
    my $target;
    my $uri = '';
    OUTER: for my $site (keys %$smap) {
        my $publ = $smap->{$site}->{publish};
        next unless ref $publ eq 'HASH';
        for my $label (keys %$publ) {
            $target = $publ->{$label}->{dest_web_dir};
            next if ! $target || $web_file !~ m/^$target/;
            # Match! 
            $uri = $publ->{$label}->{dest_url};
            last OUTER;
        }
    }

    # If found and $uri is defined, do the rewrite
    if ($uri && $target) {
        # Check trailing slashes on $target and $uri
        if (substr($target,-1) eq '/' && substr($uri,-1) ne '/') {
            $uri .= '/';
        }
        elsif (substr($target,-1) ne '/' && substr($uri,-1) eq '/') {
            $uri = substr($uri,0,-1);
        }
        $web_file =~ s/^$target/$uri/;
    }

    return $web_file;
}

# 
# Return the site (hashref) with the longest match against $fs_file, or undef 
#   if no match found.
#
sub getSiteHash {
    my $fs_file = shift;
    $fs_file = substr($fs_file,0,-1) if substr($fs_file,-1) eq '/';

    my $smap = loadSiteMap();
    return undef unless ref $smap eq 'HASH';
    # Process the site map entries from largest path size to smallest
    for my $site (sort { length($smap->{$Ewok::File::b}->{path}) <=> length($smap->{$Ewok::File::a}->{path}) } keys %$smap) {
        next if $fs_file !~ m/^$smap->{$site}->{path}/;
        # Match - add the site as a label in case we need it, and return
        $smap->{$site}->{label} = $site;
        return $smap->{$site};
    }
    return undef;
}

#
# Map current directory settings to publish directories 
#   $publish_dir is the local publish directory
#   $publish_url is the url where to use to access published content
#   If no dest_url is supplied, $publish_url eq $publish_dir
#
sub getPublishDirs 
{
    my ($web_dir, $src_web_dir, $dest_web_dir, $dest_url) = @_;
    $web_dir .= '/' unless substr($web_dir,-1) eq '/';

    return (undef,undef) unless $web_dir && $src_web_dir && $dest_web_dir;
    
    # Publish directories are obtained by replacing a $src_web_dir prefix
    #   with $dest_web_dir and/or $dest_url
    my ($publish_dir, $publish_url);
    ($publish_dir = $web_dir) =~ s/^$src_web_dir/$dest_web_dir/;
    if ($dest_url) {
        ($publish_url = $web_dir) =~ s|^$src_web_dir|$dest_url/|;
    }
    else {
        $publish_url = $publish_dir;
    }

    # Clean up '/'s in $publish_url
    $publish_url =~ s|//+|/|g;
    $publish_url =~ s|:/|://|g;
    $publish_url =~ s|/$||;

    return ($publish_url, $publish_dir);
}

#
# Delete the files in @$filelist from the publish directory for $fs_dir.
#
sub unpublish ($$$;$) 
{
    my ($req, $fs_dir, $filelist, $arg) = @_;
    my $regex = $arg->{regex};
    my $web_dir = $arg->{web_dir};
    my $site = $arg->{site};
    my $errors = $arg->{errors} || [];
    my $errorCount = scalar @$errors;

    # Determine site if not passed in
    $site ||= getSiteHash($fs_dir);

    my ($src_fs_dir, $dest_fs_dir);
    if (ref $site eq 'HASH' && ref $site->{publish} eq 'HASH') {
        $src_fs_dir  = $site->{alias};
        my $target = $site->{default_target} || (sort keys %{$site->{publish}})[0];
        $dest_fs_dir = $site->{publish}->{$target} 
            if exists $site->{publish}->{$target};
    }

    # Check for required parameters
    warning("Error: no files given to unpublish.", $errors) if ! @$filelist;
    warning("Error: publish source directory not set.", $errors) if ! $src_fs_dir;
    warning("Error: publish destination directory not set.", $errors) if ! $dest_fs_dir;
    return undef if scalar(@$errors) > $errorCount;

    # Check the top-level target directory exists
    my $fs_pub_dir = "$fs_dir/";
    $fs_pub_dir =~ s/^$src_fs_dir/$dest_fs_dir/;
    $fs_pub_dir =~ s|/+$||;
    if (! -d $fs_pub_dir) {
        warning("Error: publish directory $fs_pub_dir not found.", $errors);
        return undef;
    }

    # Unlink the given files if $fs_dir/$f does not exist
    for my $f (@$filelist) {
        next if $regex && $f !~ m/$regex/;
        next if -e "$fs_dir/$f";
        my $f = $fs_pub_dir . $f;
        if (-f $f || -l $f) {
            unlink $f;
        }
        # TODO: directories here!
    }

    # Do post_publish, if set
#   if ($EWOK_ALLOW_EXEC && $pubHash->{post_publish}) {
#       my $out = `$pubHash->{post_publish}`;
#       info("post_publish: $out");
#   }

    return 1;
}

#
# Delete any files that exist in $fs_pub_dir that do not exist in $fs_dir.
#
sub syncDeletes ($$$;$)
{
    my ($req, $fs_dir, $fs_pub_dir, $arg) = @_;
    my $test = $arg->{test};
    my ($d1, $f1) = Ewok::File::readdir($fs_dir);
    my ($d2, $f2) = Ewok::File::readdir($fs_pub_dir);
    
    my @deleted = ();
    for my $f (@$f2) {
        ($f) = (uri_unescape($f) =~ m|.*?([^/]+)$|);
        if (! grep /^$f$/, @$f1) {
            push @deleted, $f;
            unlink "$fs_pub_dir/$f" if ! $test;
        }
    }

    return @deleted;
}

#
# For EMBPERL_FILESMATCH files, use EmbperlObject::Execute to publish. 
#   Used for standard files, and for links if $copylinks is set.
#
sub publish_embperl
{
    my ($fs_dir, $f, $fs_pub_dir, $f2, $arg) = @_;
    my $force = $arg->{force};
    my $test = $arg->{test};
    my $req = $arg->{req};
    my $results = $arg->{results};
    my $errors = $arg->{errors};
    my $web_dir = $arg->{web_dir};
    my $target = $arg->{target};
    my $template = $arg->{template};
    my $dest_url = $arg->{dest_url};
    my $dest_host = $arg->{dest_host};
    my $src_fs_dir = $arg->{src_fs_dir};
    my $src_web_dir = $arg->{src_web_dir};
    my $dest_fs_dir = $arg->{dest_fs_dir};
    my $dest_web_dir = $arg->{dest_web_dir};
    my $template_dir = $arg->{template_dir};
    $template_dir .= ';' if $template_dir;
    my $Config = get_config();
    my $debug = $Config->{EWOK_ERROR_LEVEL} eq 'debug';

    my ($old, $new);
    $ENV{EWOK_PUBLISHED} = 1;
    $ENV{SCRIPT_NAME} = "$web_dir/$f";
    my $fdat = { t => $template } if $template;
    my $options = $ENV{EMBPERL_OPTIONS};
    my $addpath = "$template_dir$EWOK_HOME/htdocs/templates;$EWOK_HOME/htdocs";
    $addpath =~ s!//+!/!g;
    debug("[publish_embperl] execute: inputfile '$fs_dir/$f', template '$template', options '$options', object_base '$ENV{EMBPERL_OBJECT_BASE}', object_stopdir '$src_fs_dir', object_addpath '$addpath'") if $debug;
    my $execute = {
        inputfile => "$fs_dir/$f",
        output => \$new,
        fdat => $fdat,
        options => $options,
        object_base => $ENV{EMBPERL_OBJECT_BASE},
        object_stopdir => $src_fs_dir,
        object_addpath => $addpath,
    };
    if ($Config->{EMBPERL_VERSION} == 2) {
        require EmbperlObject;
        EmbperlObject::Execute($execute);
    }
    else {
        require HTML::EmbperlObject;
        HTML::EmbperlObject::Execute($execute);
    }
    # Trim leading and trailing spaces
    $new =~ s/^\s+//s;
    $new =~ s/\s+$/\n/s;
    # If no output, push error and return
    if ($new eq '' || $new eq "\n") {
        warning("Error: publishing file '$f' produced no output - not published.", $errors);
        return undef;
    }
    my $newsize1 = length($new);

    # Do URL rewrites if EWOK_PUBLISH_REWRITES set
    if ($Config->{EWOK_PUBLISH_REWRITES} && $target ne 'stage') {
        # By default, only rewrite urls for the current site, for performance
#       if (! $EWOK_PUBLISH_REWRITE_ALL_SITES) {
            my $from = $src_web_dir;
            $from .= '/' unless substr($from,-1) eq '/';
            my $to = $dest_url;
            if ($to) {
                # Remove host part if matches $dest_host
                $to =~ s|^$dest_host|/|;
            }
            else {
                $to = $dest_web_dir;
            }
            debug("[publish_embperl] rewriting from $from, to $to, dest_host $dest_host, dest_url $dest_url") if $debug;
            $new =~ s/(\b(href|src)\s*=\s*")$from/$1$to/ig;
            $new =~ s/(\burl\s*\(\s*)$from/$1$to/ig;
            my $host = $dest_host;
            $host =~ s|/+$||;
            $new =~ s|<!--\s*EWOK_REWRITE\s*-->\s*(.*)http://$ENV{HTTP_HOST}$from|$1$host$to|ig;
#       }
        # Rewrite URLs for ALL ewok sites
#       else {
#           $site ||= getSiteHash($fs_dir);
#           # Iterate in reverse key length order through %$EWOK_PUBLISH_MAP
#           # This should maybe be reverse src_web_dir order?
#           for my $key (reverse sort {length $a <=> length $b} keys %$EWOK_PUBLISH_MAP) {
#               my $pubHash = getPublishHash($key, $target);
#               next if ! defined $pubHash;
#               my $from = $pubHash->{src_web_dir};
#               my $to = $pubHash->{dest_url};
#               if ($to) {
#                   # Remove host part if matches $dest_host
#                   $to =~ s|^$dest_host|/|;
#               }
#               else {
#                   $to = $pubHash->{dest_web_dir};
#               }
#               if (! $from || ! $to) {
#                   if (! $from) {
#                       warning("Error: unable to derive sensible 'from' component from '$pubHash->{src_web_dir}' for URL rewriting.\n", $errors);
#                   }
#                   if (! $to) {
#                       warning("Error: unable to derive sensible 'to' component from '$pubHash->{dest_web_dir}' for URL rewriting.\n", $errors);
#                   }
#                   return undef;
#               }
#               $new =~ s/(\b(href|src)\s*=\s*")$from/$1$to/ig;
#               $new =~ s/(\burl\s*\(\s*)$from/$1$to/ig;
#               my $host = $dest_host;
#               $host =~ s|/+$||;
#               $new =~ s|<!--\s*EWOK_REWRITE\s*-->\s*(.*)http://$ENV{HTTP_HOST}$from|$1$host$to|ig;
#           }
#       }
    }
    if ($debug) {
        my $old = -s "$fs_dir/$f";
        debug("[publish_embperl] $fs_dir/$f publish sizes: old $old, new initial $newsize1, new url rewrites " . length($new));
    }

    # Check there are differences if previously published file exists
    if (! $force && -f "$fs_pub_dir/$f2" && open(OLD, "<", "$fs_pub_dir/$f2")) {
        {
          local $/ = undef;
          $old = <OLD>;
        }
        close OLD;
        # Trim leading and trailing spaces
        $old =~ s/^\s*//so;
        $old =~ s/\s*$/\n/so;
        # Check for differences
        if (! $old) {
            debug("[publish_embperl] old publish file $fs_pub_dir/$f2 apparently empty");
        }
        elsif ($new eq $old) {
            push @{$results->{warnings}}, "Warning: file '$f2' identical to current published version - not published.";
            debug("[publish_embperl] publish file $fs_pub_dir/$f2 unchanged - not publishing.");
            return 0;
        }
        else {
            debug("[publish_embperl] publish file $fs_pub_dir/$f2 differs from previous version - publishing");
        }
    }

    # Publish - do the update if not in test mode
    if (! $test) {
        debug("[publish_embperl] doing write to $fs_pub_dir/$f2");
        # Check whether $f2 contains any further subdirectories
        if ($f2 =~ m|/|) {
            my $subdir = $f2;
            $subdir =~ s|/[^/]+$||;
            if (! -d "$fs_pub_dir/$subdir") {
                if (! makedir("$fs_pub_dir/$subdir", { check_dir => $dest_fs_dir, errors => $errors })) {
                    debug("[publish_embperl] makedir of '$fs_pub_dir/$subdir' failed.", @$errors);
                    return undef;
                }
            }
        }
        # If target is a link, remove it explicitly
        unlink "$fs_pub_dir/$f2" if -l "$fs_pub_dir/$f2";
        # Write $new out to a file
        if (! open NEW, ">", "$fs_pub_dir/$f2") {
            warning("Error: opening file '$f2' for writing failed - not published.", $!, $errors);
            return undef;
        }
        elsif (! print NEW $new) {
            warning("Error: writing to file '$f2' failed.", $!, $errors);
            close NEW;
            return undef;
        }
        close NEW;
    }
    else {
        debug("[publish_embperl] test mode - skipping write of $fs_pub_dir/$f2");
    }

    return 1;
}

sub publish_link
{
    my ($fs_dir, $f, $fs_pub_dir, $f2, $arg) = @_;
    my $force = $arg->{force};
    my $test = $arg->{test};
    my $results = $arg->{results};
    my $errors = $arg->{errors};

    # For links, recreate in publish area (assumes *always* relative)
    my ($target) = (readlink("$fs_dir/$f") =~ m|(\S.*?)\s*$|);   #/
    if (! $force && -l "$fs_pub_dir/$f2" &&
            readlink("$fs_pub_dir/$f2") eq $target) {
        push @{$results->{warnings}}, "Warning: link '$f' identical to current published version - not published.";
        return 0;
    }
    elsif (! $test) {
        if (-f "$fs_pub_dir/$f2" || -l "$fs_pub_dir/$f2") {
            unlink "$fs_pub_dir/$f2"; 
        }
        elsif (-d "$fs_pub_dir/$f2") {
            info("Error: publish of link '$f' failed - '$f2' already exists as directory.", $errors);
            return undef;
        }
        if (! chdir $fs_pub_dir) {
            info("Error: chdir to publish directory failed.", $!, $errors);
            return undef;
        }
        if (!symlink $target, "$fs_pub_dir/$f2") {
            info("Error: publish of link '$f' failed.", $!, $errors);
            return undef;
        }
    }
    return 1;
}

#
# For non-EMBPERL_FILESMATCH files, simply publish as copy
#
sub publish_copy
{
    my ($fs_dir, $f, $fs_pub_dir, $f2, $arg) = @_;
    my $force = $arg->{force};
    my $test = $arg->{test};
    my $results = $arg->{results};
    my $errors = $arg->{errors};
    my $dest_fs_dir = $arg->{dest_fs_dir};

    if (! $force && -f "$fs_pub_dir/$f2" && 
        compare("$fs_dir/$f","$fs_pub_dir/$f2") == 0) {
        push @{$results->{warnings}}, "Warning: file '$f' identical to current published version - not published." if $f !~ m/.htaccess$/;
        return 0;
    }

    # Copy to $fs_pub_dir
    elsif (! $test) {
        if ($f2 =~ m|/|) {
            my $subdir = $f2;
            $subdir =~ s|/[^/]+$||;
            if (! -d "$fs_pub_dir/$subdir") {
                if (! makedir("$fs_pub_dir/$subdir", { check_dir => $dest_fs_dir, errors => $errors })) {
                    return undef;
                }
            }
        }
        # If target is a link, remove it explicitly
        unlink "$fs_pub_dir/$f2" if -l "$fs_pub_dir/$f2";
        if (! copy("$fs_dir/$f", "$fs_pub_dir/$f2")) {
            info("Error: publish (copy) of file '$f' failed.", $!, $errors);
            return undef;
        }
    }

    return 1;
}

#
# Publish the given $filelist from staging (the $src_fs_dir tree) to
#   production ($dest_fs_dir tree).
# Return undef if errors, true otherwise. Results are returned in the 
#   $results hashref, typically 'pubList', 'deletes', 'errors', and
#   'warnings'.
#
sub publish
{
    my $req = shift;
    my $filelist = shift;
    my $fs_dir = shift;
    my $results = shift;
    my %arg = @_;
    my $web_dir = $arg{web_dir};
    my $refresh = $arg{refresh};
    my $target = $arg{target};
    my $template = $arg{template};
    my $pubfilename = $arg{pubfilename};
    my $stage_replace = $arg{stage_replace};
    my $regex = $arg{regex};
    my $test = $arg{test};
    my $force = $arg{force};
    my $label = $arg{label};
    my $site = $arg{site};
    my $fdat = $arg{fdat};
    my $copylinks = $arg{copylinks};
    my $sync_deletes = $arg{sync_deletes};
    my $scope = $arg{scope} || 'selected';
    my $debug = $arg{debug};

    $filelist = [ $filelist ] if ref $filelist ne 'ARRAY';
    $web_dir = $fs_dir if ! defined $web_dir;

    # Remove any trailing slashes
    $fs_dir  =~ s|/+$||;
    $web_dir =~ s|/+$||;

    # Untaint
    ($target) = ($target =~ m/(\S.*?)\s*$/);

    # Errors, warnings, and pubList can be ignored if not passed as arrayrefs
    $results->{pubList} = [] if ref $results->{pubList} ne 'ARRAY';
    $results->{deleteList} = [] if ref $results->{deleteList} ne 'ARRAY';
#   $results->{errors} = [] if ref $results->{errors} ne 'ARRAY';
    $results->{warnings} = [] if ref $results->{warnings} ne 'ARRAY';
    my $errors = $results->{errors};
    my $errorCount = scalar @$errors;

    # Determine site if not passed in
    $site ||= getSiteHash($fs_dir);

    # Setup publish directories
    my $src_fs_dir = $site->{path};
    my $src_web_dir = $site->{alias};
    $template ||= $site->{publish}->{$target}->{template};
    my $template_dir = $site->{publish}->{$target}->{template_dir} ||
        "$src_fs_dir/templates";
    my ($dest_web_dir, $dest_fs_dir, $dest_host, $dest_url);
    if ($target eq 'stage') {
        $dest_fs_dir = $src_fs_dir;
        $dest_web_dir = $web_dir;
    }
    else {
        $dest_fs_dir = $site->{publish}->{$target}->{dest_fs_dir};
        $dest_web_dir = $site->{publish}->{$target}->{dest_web_dir};
        $dest_url = $dest_host = $site->{publish}->{$target}->{dest_url};
        $dest_host =~ s|^(https?://[^/]+/).*$|$1|;
    }

    # Check for required parameters
    info("Error: no files given to publish.", $errors)
        if ! @$filelist && ! $sync_deletes;
    info("Error: publish source directory not set.", $errors)
        if ! $src_fs_dir;
    info("Error: publish destination directory not set.", $errors)
        if ! $dest_fs_dir;
    # Silently prune illegal files for 'dir', 'tree', and 'refresh' publishes
    my @prune = ( prune => 1 ) if $scope ne 'selected' || $refresh;
    restrictCheck($req, 'publish', $src_fs_dir, $filelist, { errors => $errors, @prune, web_dir => $web_dir });
    return undef if scalar(@$errors) > $errorCount;

    # Check we are publishing from within the $src_fs_dir tree
    warning("Error: directory '$web_dir' not within publish source tree '$src_fs_dir'.", $errors)
        if ("$fs_dir/" !~ m/^$src_fs_dir/);
    return undef if scalar(@$errors) > $errorCount;

    # Check the top-level target directory exists
    my $fs_pub_dir = "$fs_dir/";
    $fs_pub_dir =~ s/^$src_fs_dir/$dest_fs_dir/ if $target ne 'stage';
    $fs_pub_dir =~ s|/+$||;
    return undef if ! makedir($fs_pub_dir, { check_dir => $dest_fs_dir, 
        errors => $errors, copylinks => $copylinks });

    # Setup logger
#   my $t = Ewok::Logger->new('File::publish');
#   $t->info("[publish] $req->{user}: $web_dir($fs_dir), " . join(':',@$filelist) . ", $sync_deletes");

    if (@$filelist) {
        # Checkpoint globals that can get clobbered on publish
        executeCheckpoint($req, $fdat);

        # Publish the individual files
        for my $f (@$filelist) {
            # Untaint
            ($f) = ($f =~ m|^\s*(.+)\s*$|); 
            # Check for print version
            my $meta = getMeta("$fs_dir/$f");
            my $fp;
            if ($meta->{print_version}) {
                ($fp = $f) =~ s/(\.[^\.]*)$/.print$1/;
            }

            # Setup of publish filename
            my $f2 = $pubfilename;
            if ($f2) {
                $f2 .= $f if -d "$fs_pub_dir/$f2";
            }
            else {
                $f2 = $f;
            }
            return undef if $regex && $f !~ m/$regex/;
            if (! -e "$fs_dir/$f") {
                info("Error: '$f' does not exist - not published.", $errors)
            }
            elsif (! -f "$fs_dir/$f" && ! -l "$fs_dir/$f") {
                info("Error: '$f' is not a file - not published.", $errors)
            }
            elsif ($f =~ m/\.epl$/i) {
                info("Warning: cannot publish '.epl' files - '$f' not published.", $results->{warnings});
            }

            next if scalar(@$errors) > $errorCount;

            if ($target eq 'stage' && -e "$fs_pub_dir/$f2" && ! $stage_replace) {
                info("Error: '$f2' already exists - select 'Force replace' option to overwrite.", $errors)
            }
            # Refresh only if existing
            elsif ($refresh && ! -e "$fs_pub_dir/$f2") {
                debug("refresh mode and '$fs_pub_dir/$f2' does not exist - not published");
                next;
            }

            # Do the actual publish
            my $publish = 0;
            if ($f =~ $ENV{EMBPERL_FILESMATCH} &&
              (! -l "$fs_dir/$f" || $copylinks)) {
                $publish = publish_embperl($fs_dir, $f, $fs_pub_dir, $f2,
                    { test => $test, refresh => $refresh, 
                      req => $req, force => $force, 
                      errors => $errors, results => $results, 
                      web_dir => $web_dir, target => $target, 
                      template => $template, dest_host => $dest_host, 
                      dest_url => $dest_url, dest_web_dir => $dest_web_dir,
                      src_fs_dir => $src_fs_dir, src_web_dir => $src_web_dir,
                      dest_fs_dir => $dest_fs_dir,
                      template_dir => $template_dir, debug => $debug, });
                # Publish print version, if applicable
                if ($fp && $template ne 'print') {
                    my $print = publish_embperl($fs_dir, $f, $fs_pub_dir, $fp,
                        { test => $test, refresh => $refresh, 
                          req => $req, force => $force,
                          errors => $errors, results => $results, 
                          web_dir => $web_dir, target => $target, 
                          template => 'print', dest_host => $dest_host, 
                          dest_url => $dest_url, dest_web_dir => $dest_web_dir,
                          src_fs_dir => $src_fs_dir, 
                          src_web_dir => $src_web_dir,
                          dest_fs_dir => $dest_fs_dir,
                          template_dir => $template_dir, debug => $debug, });
                    push @{$results->{pubList}}, $fp 
                        if $print && $f !~ m/.htaccess$/;
                }
            }
            elsif (-l "$fs_dir/$f" && !$copylinks) {
                $publish = publish_link($fs_dir, $f, $fs_pub_dir, $f2,
                    { test => $test, refresh => $refresh, 
                      force => $force, debug => $debug,
                      errors => $errors, results => $results, });
            }
            else {
                $publish = publish_copy($fs_dir, $f, $fs_pub_dir, $f2,
                    { test => $test, refresh => $refresh, 
                      force => $force, debug => $debug,
                      errors => $errors, results => $results, 
                      dest_fs_dir => $dest_fs_dir, });
            }

            # Success - add file to pubList
            push @{$results->{pubList}}, $f2 if $publish && $f !~ m/.htaccess$/;
        }

        # Restore those clobberable globals
        executeRestore($req, $fdat);
    }

    # Do syncDeletes
    push @{$results->{deleteList}}, syncDeletes($req, $fs_dir, $fs_pub_dir,
         { test => $test }) if $sync_deletes;

    # Do a 'publish' authzSave
    my $Config = get_config();
    if ($Config->{EWOK_AUTHZ_PUBLISH} && 
      -f "$fs_dir/$Config->{EWOK_AUTHZ_OBJECT}") {
        if (!authzSave($req, $fs_pub_dir, authzLoad($fs_dir), 
                { publish => $target, label => $label })) {
            warning("Error saving permissions object.", $!, $errors)
        }
    }

    # Do post_publish, if set
#   if ($EWOK_ALLOW_EXEC && $pubHash->{post_publish} && ! $test) {
#       my $out = `$pubHash->{post_publish}`;
#       warning("post_publish: $out") if $out;
#   }

    # Return true if no errors
    return ! scalar(@$errors) > $errorCount;
}

#
# Wrapper subroutine generating all the files to publish at or below the $fs_dir.
#   Calls &publish with this list.
#
sub publishAll
{
    my $req = shift;
    my $fs_dir = shift;
    my $results = shift;
    my %arg = @_;
    my $errors = $results->{errors};
    my $recursive = $arg{recursive};
    my $copylinks = $arg{copylinks};
    my $web_dir = $arg{web_dir};
    $arg{scope} = $recursive ? 'tree' : 'directory';

    $recursive = 0 if ! defined $recursive;
    $web_dir = $fs_dir if ! defined $web_dir;
    $errors = [] if ref $errors ne 'ARRAY';

    # Remove any trailing slashes
    $fs_dir  =~ s|/+$||;
    $web_dir =~ s|/+$||;
 
    # Check whether this directory is actually a link; if so, dereference it
    my $fs_link;
    if (-l $fs_dir && ! $copylinks) {
        $fs_link = $fs_dir;
        my $linkdir = readlink($fs_dir);
        # For relative links, need to munge the filename
        if (substr($linkdir,0,1) ne '/') {
            $fs_dir =~ s|[^/]+/*$||;
            $fs_dir .= "/$linkdir";
            while ($fs_dir =~ s|/+[^\./]+/+\.\.||g) { 1 };
        }
    }
 
    my @dirlist = ( $fs_dir );
    my @filelist = ();
    while (@dirlist) {
        my $dir = shift @dirlist;
        my $rel_dir = $dir;
        $rel_dir =~ s|^$fs_dir||;
        $rel_dir = substr($rel_dir,1) if substr($rel_dir, 0, 1) eq '/';
        if (! -d "$dir") {
            warning("Error: directory $web_dir/$rel_dir not found.", $errors);
            next;
        }
        if (!opendir DIR, "$dir") {
            warning("Error: unable to open directory $web_dir/$rel_dir.", $errors);
            next;
        }
        my @list = reverse sort(CORE::readdir(DIR));
        for (@list) {
            next if m/^\./;
            # If recursive (tree scope) add directories to dirlist
            if (-d "$dir/$_") {
                # Add to dirlist if recursive && ! source-control dir
                unshift @dirlist, "$dir/$_" if $recursive && ! m/^(CVS|RCS|{arch})$/;
                next;
            }
            # TODO: need a check for locked files, presumably omitted
            push @filelist, (($rel_dir) ? "$rel_dir/$_" : $_);
        }
        closedir DIR;
    }
 
    if (@filelist) {
        # If this is a directory link, publish the link itself as well
        if ($fs_link) {
            $fs_link =~ m|^(.*/)([^/]+/*$)|;
            # Copy the publishargs and explicitly turn off copylinks
            # my $fs_arg = { map { $_ => $arg{$_} } keys %arg };
            # $fs_arg->{copylinks} = 0;
            my $pubList = $results->{pubList};
            $results->{pubList} = [];
            publish($req, [ $2 ], $1, $results, %arg) if $1 && $2;
            # Fake a null results filename for the link itself
            push @$pubList, '';
            $results->{pubList} = $pubList;
        }
        # Call publish with the filelist
        return publish($req, \@filelist, $fs_dir, $results, %arg);
    }
}
 
1;

__END__

=head1 NAME

Ewok::Publish - ewok publishing operations


=head1 SYNOPSIS

    # TODO


=head1 DESCRIPTION

Ewok::Publish provides various publishing routines.


=head1 AUTHOR

Gavin Carr <gavin@openfusion.com.au>


=head1 COPYRIGHT

Copyright 2003, Open Fusion Pty. Ltd., Australia. All Rights Reserved.

This program is free software. You may copy or redistribute it under the same 
terms as perl itself.

=cut
